#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / cache / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'cache_hub',
        ]

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry

CACHE_DRIVERS_KEY = 'drivers'
CACHE_CONF_KEY = 'caches'
DEFAULT_CACHE_NAME = 'main'

PROBER = ModProber('weiyu.cache.drivers', '%s_driver')


class CacheHub(BaseHub):
    registry_name = 'weiyu.cache'
    registry_class = UnicodeRegistry
    handlers_key = CACHE_DRIVERS_KEY

    def __init__(self):
        super(CacheHub, self).__init__()

        self._cfg = {}

    def get_cache(self, name=DEFAULT_CACHE_NAME):
        try:
            drv, opts = self._cfg[name]
        except KeyError:
            try:
                opts = self._reg[CACHE_CONF_KEY][name].copy()
            except KeyError:
                raise ValueError("cache '%s' not configured" % (name, ))

            drv = opts.pop('driver')
            self._cfg[name] = (drv, opts, )

            # Load the cache backend
            PROBER.modprobe(drv)

        return self.do_handling(drv, opts)


cache_hub = CacheHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / http / utilities
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'status_to_str',
            'dummy_file_wrapper',
            'send_content_iter',
            'parse_form',
            'gen_http_headers',
            ]

from functools import partial

try:
    # Python 3
    from urllib.parse import parse_qs
except ImportError:
    # Python 2
    from urlparse import parse_qs

try:
    import ujson as json
except ImportError:
    import json

import six

from ...helpers.misc import smartbytes


# This dict is pasted from Django's core/handlers/wsgi.py
# See http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
STATUS_CODES_MAP = {
        100: 'CONTINUE',
        101: 'SWITCHING PROTOCOLS',
        200: 'OK',
        201: 'CREATED',
        202: 'ACCEPTED',
        203: 'NON-AUTHORITATIVE INFORMATION',
        204: 'NO CONTENT',
        205: 'RESET CONTENT',
        206: 'PARTIAL CONTENT',
        300: 'MULTIPLE CHOICES',
        301: 'MOVED PERMANENTLY',
        302: 'FOUND',
        303: 'SEE OTHER',
        304: 'NOT MODIFIED',
        305: 'USE PROXY',
        306: 'RESERVED',
        307: 'TEMPORARY REDIRECT',
        400: 'BAD REQUEST',
        401: 'UNAUTHORIZED',
        402: 'PAYMENT REQUIRED',
        403: 'FORBIDDEN',
        404: 'NOT FOUND',
        405: 'METHOD NOT ALLOWED',
        406: 'NOT ACCEPTABLE',
        407: 'PROXY AUTHENTICATION REQUIRED',
        408: 'REQUEST TIMEOUT',
        409: 'CONFLICT',
        410: 'GONE',
        411: 'LENGTH REQUIRED',
        412: 'PRECONDITION FAILED',
        413: 'REQUEST ENTITY TOO LARGE',
        414: 'REQUEST-URI TOO LONG',
        415: 'UNSUPPORTED MEDIA TYPE',
        416: 'REQUESTED RANGE NOT SATISFIABLE',
        417: 'EXPECTATION FAILED',
        500: 'INTERNAL SERVER ERROR',
        501: 'NOT IMPLEMENTED',
        502: 'BAD GATEWAY',
        503: 'SERVICE UNAVAILABLE',
        504: 'GATEWAY TIMEOUT',
        505: 'HTTP VERSION NOT SUPPORTED',
        }

# Content types that can be parsed into a form
_FORM_CONTENT_HANDLERS = {}


def _form_content_handler(content_type):
    def _decorator_(fn):
        if content_type in _FORM_CONTENT_HANDLERS:
            raise ValueError(
                    "'%s' already registered as handler!" % (
                        content_type,
                        )
                    )

        _FORM_CONTENT_HANDLERS[content_type] = fn
        return fn
    return _decorator_


def status_to_str(status):
    '''Converts status code to its description.

    '''

    return STATUS_CODES_MAP[status]


def dummy_file_wrapper(fp, blk_sz=None):
    blk_sz = blk_sz if blk_sz is not None else 4096
    do_read = partial(fp.read, blk_sz)
    chunk = do_read()
    while len(chunk) > 0:
        yield chunk
        chunk = do_read()
    fp.close()


def send_content_iter(content, enc):
    if isinstance(content, six.string_types):
        yield smartbytes(content, enc, 'replace')
    elif isinstance(content, six.binary_type):
        yield content
    else:
        for chunk in content:
            # encode and send the chunk using response.encoding
            yield smartbytes(chunk, enc, 'replace')


def parse_form(ctype, content):
    if ctype in _FORM_CONTENT_HANDLERS:
        return _FORM_CONTENT_HANDLERS[ctype](content)

    return None


@_form_content_handler('application/x-www-form-urlencoded')
def _parse_urlencoded_form(content):
    form = parse_qs(content)

    # eliminate all those 1-element lists
    for k in six.iterkeys(form):
        if len(form[k]) == 1:
            form[k] = form[k][0]

    return form


@_form_content_handler('application/json')
def _parse_json_form(content):
    return json.loads(content)


def gen_http_headers(response, __Content_Length=str('Content-Length')):
    status_code, enc = response.status, response.encoding

    status_line = str('%d %s' % (
            status_code,
            status_to_str(status_code),
            ))

    # ensure all header contents are bytes
    headers = []

    # insert a Content-Length along if response is not raw file and contains
    # a body
    if not response.is_raw_file and not response._dont_render:
        headers.append((
                __Content_Length,
                str(len(response.content)),
                ))

    for k, v in response.http_headers:
        k_bytes, v_bytes = smartbytes(k, enc), smartbytes(v, enc)
        k_str, v_str = str(k_bytes), str(v_bytes)
        headers.append((k_str, v_str, ))

    return status_line, headers


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / session handling / Redis-based sessions
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'RedisSessionObject',
        'RedisSession',
        ]

import time
import uuid

import six
import redis

# I don't feel like reinventing the wheel, so...
import Cookie

from beaker.session import SignedCookie

from ..db import db_hub
from . import session_hub


def _generate_sessid():
    return uuid.uuid4().hex


class RedisSessionObject(dict):
    def __init__(
            self,
            sid,
            id=None,
            cookiehdr=None,
            key=None,
            secret=None,
            ttl=None,
            ):
        self.sid = sid
        self.id, self.key, self.secret, self.ttl = id, key, secret, ttl
        self.cookiehdr = cookiehdr
        self.cookie, self._do_set_cookie = None, False

        # first get the appropriate key for db object, generating one if
        # the session is new.
        if cookiehdr is not None:
            # session ID stored in cookie. the code below is again taken
            # from beaker
            if secret is not None:
                try:
                    cookie = SignedCookie(secret, input=cookiehdr)
                except Cookie.CookieError:
                    cookie = SignedCookie(secret, input=None)
            else:
                cookie = Cookie.SimpleCookie(input=cookiehdr)

            if self.id is None and key in cookie:
                self.id = cookie[key].value

            self.cookie = cookie

        if self.id is None:
            # Generate a new ID, and (indirectly) instruct the session
            # middleware to Set-Cookie.
            self.new_id()

        self._drv = db_hub.get_storage(sid)
        # Unlike Beaker, which deals with a file backend without automatic
        # expiration of entries, we don't have to do all that ourselves.
        # Just making sure an EXPIRE is issued for (theoretically) each
        # operation would be enough...
        #
        # if only all the keys are pre-existing. Actually, hashes are
        # created only when they are written to, so we have no choice but
        # to also force an EXPIRE every time a write is issued.
        with self._drv as conn:
            if conn.exists(self.id):
                conn.expire(self.id, self.ttl)
            else:
                self.new_id()

    # helper methods
    def new_id(self):
        old_id = self.id
        self.id, self._do_set_cookie = _generate_sessid(), True

        # migrate the session in DB
        if old_id is not None:
            with self._drv as conn:
                old_content = conn.hgetall(old_id)

                # A hash cannot be empty in Redis, so if it indeed is
                # empty it must not exist.
                if old_content:
                    # the probability of UUID conflicts is ~0,
                    # so why do we have to say while True here?
                    result = False
                    for i in xrange(4):
                        try:
                            result = conn.renamenx(old_id, self.id)
                            if result:
                                # rename successful
                                break
                            # rename failed because target key exists
                            self.id = _generate_sessid()
                        except redis.exceptions.ResponseError:
                            # rename failed because source and destination
                            # are the SAME... or because the source has
                            # VANISHED in operation!
                            self.id = _generate_sessid()

                            if not conn.exists(old_id):
                                # well, the session expired during <<1ms of
                                # execution.
                                # Lucky we have old_content... Just assign it
                                # to self.id. Another edge case is that the
                                # newly-generated id happens to conflict with
                                # something else... Oh well. Let's not go that
                                # far.
                                conn.hmset(self.id, old_content)
                                result = True
                                break

                    if not result:
                        # That many *UUID conflicts* ?
                        # Better go buy some lotteries...
                        raise RuntimeError('Session UUID conflicts!')

        self.set_cookie_prop()

    def _full_refresh(self):
        s = super(RedisSessionObject, self)

        # Single-direction sync from redis.
        # In the rare case of incorrect object type (should never be the
        # case if this code is the only one accessing that db), data stored
        # in the Python dict will be dropped, then the exception is
        # re-raised.
        with self._drv as conn:
            try:
                new_self = conn.hgetall(self.id)
            except redis.exceptions.ResponseError:
                s.clear()
                raise

        s.clear()
        s.update(new_self)

    def __repr__(self):
        return b'RedisSessionObject(%s)' % (
                super(RedisSessionObject, self).__repr__(),
                )

    # Cookie operations
    def set_cookie_prop(self, expires=None, domain=None, path='/'):
        self.cookie[self.key] = self.id
        entry = self.cookie[self.key]

        entry['path'] = path
        if domain is not None:
            entry['domain'] = domain

        if expires is not None:
            # NOTE: assume it's in seconds for the moment
            expire_time = time.gmtime(int(time.time()) + expires)
            expire_str = time.strftime(
                    '%a, %d-%b-%Y %H:%M:%S GMT',
                    expire_time,
                    )
            entry['expires'] = expire_str

        self._do_set_cookie = True

    def generate_cookie_header(self):
        if not self._do_set_cookie:
            # the cookie doesn't want to be refreshed
            return False, ''

        return True, self.cookie[self.key].output(header='')

    # wrappers around the dict interface
    # Operations on non-existent hashes either return None or [], so we
    # don't have to care about the hash's existence. Instead we can make
    # something that wraps Redis in a *really* thin manner...
    def __len__(self):
        with self._drv as conn:
            return conn.hlen(self.id)

    def __iter__(self):
        self._full_refresh()
        return super(RedisSessionObject, self).__iter__()

    def __getitem__(self, key):
        # Fetch the key from db.
        # The reason for not caching queried fields, is possible race
        # conditions with another client which attempts to update the same
        # session, and code complexity.
        # Also, redis makes the overhead relatively bearable, so hopefully
        # performance bottleneck would not occur here.
        with self._drv as conn:
            value = conn.hget(self.id, key)

        super(RedisSessionObject, self).__setitem__(key, value)
        return value

    def __setitem__(self, key, value):
        with self._drv as conn:
            conn.hset(self.id, key, value)
            conn.expire(self.id, self.ttl)

        super(RedisSessionObject, self).__setitem__(key, value)

    def __delitem__(self, key):
        with self._drv as conn:
            conn.hdel(self.id, key)
            conn.expire(self.id, self.ttl)

        try:
            super(RedisSessionObject, self).__delitem__(key)
        except KeyError:
            pass

    def __contains__(self, key):
        with self._drv as conn:
            existence = conn.hexists(self.id, key)
            if not existence:
                try:
                    super(RedisSessionObject, self).__delitem__(key)
                except KeyError:
                    pass

            return existence

    def clear(self):
        with self._drv as conn:
            conn.delete(self.id)

        super(RedisSessionObject, self).clear()

    def get(self, key, default=None):
        return self[key] if key in self else default

    def keys(self):
        # Not HKEYS, but forces a full refresh.
        self._full_refresh()
        return six.iterkeys(super(RedisSessionObject, self))

    def iterkeys(self):
        return self.keys()

    def values(self):
        # Not HVALS, but forces a full refresh.
        self._full_refresh()
        return six.itervalues(super(RedisSessionObject, self))

    def itervalues(self):
        return self.values()

    def items(self):
        self._full_refresh()
        return six.iteritems(super(RedisSessionObject, self))

    def iteritems(self):
        return self.items()

    def setdefault(self, key, default=None):
        with self._drv as conn:
            ret = conn.hsetnx(self.id, key, default)
            conn.expire(self.id, self.ttl)
            if not ret:
                # key already exists
                value = conn.hget(self.id, key)
                super(RedisSessionObject, self).__setitem__(key, value)
                return value

            super(RedisSessionObject, self).__setitem__(key, default)
            return default

    def update(self, obj, **kwargs):
        upd_dict = obj if hasattr(obj, 'keys') else dict(obj)
        upd_dict.update(kwargs)
        with self._drv as conn:
            conn.hmset(self.id, upd_dict)
            conn.expire(self.id, self.ttl)

        super(RedisSessionObject, self).update(upd_dict)

    # useful atomic operations
    def hincrby(self, key, incr=1):
        with self._drv as conn:
            newval = conn.hincrby(self.id, key, incr)
            conn.expire(self.id, self.ttl)
            super(RedisSessionObject, self).__setitem__(key, newval)
            return newval


class RedisSession(object):
    def __init__(self, options=None):
        self.sid = options['struct_id']
        self.key = options['cookie-key']
        self.ttl = options['ttl']
        self.secret = options.get('secret', None)

    def preprocess(self, request):
        request.session = RedisSessionObject(
                self.sid,
                None,
                request.env.get('HTTP_COOKIE', ''),
                self.key,
                self.secret,
                self.ttl,
                )

    def postprocess(self, response):
        request = response.request
        ctx = response.context
        session = request.session

        set_cookie, cookiehdr = session.generate_cookie_header()
        if set_cookie:
            if cookiehdr:
                # append the cookie to response
                new_cookies = ctx.get('cookies', [])
                new_cookies.append(cookiehdr)
                ctx['cookies'] = new_cookies

        return response


@session_hub.register_handler('redis')
def redis_session_handler(hub, options=None):
    return RedisSession(options)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / helpers / metaprogramming structures

from __future__ import unicode_literals, division

__all__ = [
        'classproperty',
        'classinstancemethod',
        ]


# The following class is taken from an answer to StackOverflow question
# 3203286/how-to-create-a-read-only-class-property-in-python,
# answered by StackOverflow user bobince.
# According to SO policy, the content is licensed under cc-by-sa 3.0.
class classproperty(object):
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


# The following code is part of the formencode library, which is licensed under
# the PSF license.
# START OF formencode CODE
class classinstancemethod(object):
    """
    Acts like a class method when called from a class, like an
    instance method when called by an instance.  The method should
    take two arguments, 'self' and 'cls'; one of these will be None
    depending on how the method was called.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, type=None):
        return _methodwrapper(self.func, obj=obj, type=type)


class _methodwrapper(object):

    def __init__(self, func, obj, type):
        self.func = func
        self.obj = obj
        self.type = type

    def __call__(self, *args, **kw):
        assert 'self' not in kw and 'cls' not in kw, (
            "You cannot use 'self' or 'cls' arguments to a "
            "classinstancemethod")
        return self.func(*((self.obj, self.type) + args), **kw)

    def __repr__(self):
        if self.obj is None:
            return ('<bound class method %s.%s>'
                    % (self.type.__name__, self.func.__name__))
        else:
            return ('<bound method %s.%s of %r>'
                    % (self.type.__name__, self.func.__name__, self.obj))


# END OF formencode CODE


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / object-nonrelational mapping / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'mapper_hub',
        ]

from functools import partial

import six

from .. import db_hub
from ...helpers.hub import BaseHub
from ...registry.classes import UnicodeRegistry


DECODERS_KEY, ENCODERS_KEY = 'decoders', 'encoders'

# don't conflict w/ single-letter props to reduce number of surprises...
VERSION_FIELD = '_V'


class MapperHub(BaseHub):
    registry_name = 'weiyu.db.mapper'
    registry_class = UnicodeRegistry
    handlers_key = 'shims'

    def __init__(self):
        super(MapperHub, self).__init__()

        self._decoders = self._reg[DECODERS_KEY] = {}
        self._encoders = self._reg[ENCODERS_KEY] = {}

    def register_struct(self, name):
        if name not in self._decoders:
            self._decoders[name] = {}

        if name not in self._encoders:
            # [{ver1: encoder1, ver2: encoder2, ...}, max_version]
            self._encoders[name] = [{}, -1]

    def get_version(self, obj):
        try:
            return obj[VERSION_FIELD]
        except KeyError:
            raise TypeError(
                        'object %s has no version field' % (repr(obj), )
                        )

    def decoder_for(self, name, version):
        # you aren't registering negative versions, huh?
        assert version >= 0

        def _decorator_(fn):
            decoders = self._decoders[name]
            if version in decoders:
                raise ValueError(
                        "decoder for struct id '%s' ver %s already"
                        "exists: %s" % (
                            name,
                            six.text_type(version),
                            repr(decoders[version]),
                            )
                        )

            decoders[version] = fn
            return fn
        return _decorator_

    def encoder_for(self, name, version):
        assert version >= 0

        def _decorator_(fn):
            # update maxver if necessary
            encoders, maxver = self._encoders[name]
            if version in encoders:
                raise ValueError(
                        "encoder for struct id '%s' ver %s already"
                        "exists: %s" % (
                            name,
                            six.text_type(version),
                            repr(encoders[version]),
                            )
                        )

            encoders[version] = fn
            if version > maxver:
                self._encoders[name][1] = version

            return fn
        return _decorator_

    def decode(self, name, obj, version=None):
        decoders = self._decoders[name]

        use_ver = version if version is not None else self.get_version(obj)

        try:
            decoder = decoders[use_ver]
        except KeyError:
            raise TypeError(
                    "no decoder for struct id '%s' version %s" % (
                        name,
                        six.text_type(use_ver),
                        )
                    )

        return decoder(obj)

    def encode(self, name, obj, version=None):
        # Always encode in the latest version unless explicitly specified
        encoders, maxver = self._encoders[name]
        use_ver = maxver if version is None else version

        try:
            encoder = encoders[use_ver]
        except KeyError:
            raise TypeError(
                    "no encoder for struct id '%s' version %s" % (
                        name,
                        six.text_type(use_ver),
                        )
                    )

        result = encoder(obj)
        result[VERSION_FIELD] = maxver
        return result


mapper_hub = MapperHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / async / views
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

from socketio import socketio_manage

from . import async_hub
from ..shortcuts import http, view


@http('socketio-handler')
@view
def socketio_bridge_view(request):
    socketio_manage(
            request.env,
            async_hub.get_namespaces('socketio'),
            request,
            )

    # The handler DOES return, after finishing all operations.
    # Tell reflex to cancel further responses.
    return (
            204,
            {},
            {'request_vanished': True, },
            )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / regex based router
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'RegexRouterTarget',
            'RegexRouter',
            ]

import re

from ..helpers.misc import smartstr
from ..helpers.regex_helper import normalize

from . import router_hub
from .base import *


def args_from_match(match, named_groups):
    kwargs = match.groupdict()
    args = [i for idx, i in enumerate(match.groups())
            if idx not in named_groups
            ]
    return (args, kwargs, )


class RegexRouterTarget(RouterTargetBase):
    # __slots__ = ['name', 'pattern', 'target', '_namedgrps', ]

    def __init__(self, pattern, target, extra_data=None, router=None):
        super(RegexRouterTarget, self).__init__(target, extra_data, router)

        self.pattern = re.compile(smartstr(pattern))
        self._namedgrps = [i - 1 for i in self.pattern.groupindex.values()]

    def check(self, querystr, prev_args, prev_kwargs):
        match = self.pattern.match(smartstr(querystr))

        if not match:
            return (STATUS_NOROUTE, None, None, None, )

        curr_args, curr_kwargs = args_from_match(match, self._namedgrps)

        # update the previous context
        args = prev_args[:]
        args.extend(curr_args)

        kwargs = prev_kwargs.copy()
        kwargs.update(curr_kwargs)

        # support for nested routing
        if not self.target_is_router:
            return (STATUS_REACHED, args, kwargs, None, )

        # target is router, forward
        # since the RE matches at the beginning, safely assume its span()
        # starts at 0.
        # now, extract the substring for further routing...
        new_qs = querystr[match.span()[1]:]

        # and indicate this status.
        return (STATUS_FORWARD, args, kwargs, new_qs, )

    def _get_reverse_pattern(self):
        # self.pattern is a compiled pattern
        pat = normalize(self.pattern.pattern)
        assert len(pat) == 1
        pat_str, pat_vars = pat[0]
        return (pat_str, set(pat_vars), )


@router_hub.register_router_class('regex')
class RegexRouter(RouterBase):
    target_type = RegexRouterTarget


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / http / Tornado adapter
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'WeiyuTornadoAdapter',
            ]

try:
    from tornado.wsgi import WSGIContainer
    from tornado.httpserver import HTTPServer
except ImportError:
    # later used to check if Tornado is actually installed,
    # raising exception if it isn't the case.
    HTTPServer = object

from .. import adapter_hub

from ...helpers.misc import smartstr
from ...registry.provider import request as reg_request
from ...reflex.classes import ReflexRequest

from .base import BaseHTTPReflex
from .util import status_to_str, dummy_file_wrapper, send_content_iter
from .util import parse_form, gen_http_headers


class TornadoRequest(ReflexRequest):
    def __init__(self, tornado_req, site_conf):
        super(TornadoRequest, self).__init__(
                WSGIContainer.environ(tornado_req)
                )
        self._native_request = tornado_req

        self.site = site_conf


class TornadoReflex(BaseHTTPReflex):
    def __init__(self):
        super(TornadoReflex, self).__init__()

    def _do_accept_request(self, tornado_req):
        return TornadoRequest(tornado_req, self.SITE_CONF)

    def _do_translate_request(self, request):
        t_req = request._native_request
        # decode the path bytestring
        # TODO: improve encoding handling here
        path = request.path = smartstr(t_req.path, 'utf-8', 'replace')

        # Move routing (much) earlier so we don't waste time in processing
        # requests impossible to fulfill.
        # Note that we don't pass in "request" at this moment. The object
        # can be replaced by potential hooks, and we certainly don't want
        # a reference to be frozen in the request.
        # Return value is of format (fn, args, kwargs, route_data, )
        route_result = self._do_routing(path)
        request.callback_info = route_result[:-1]
        request.route_data = route_result[-1]

        # Rest of request object preparation goes here...
        request.remote_addr = smartstr(t_req.remote_ip)
        method = request.method = smartstr(t_req.method)
        length, content = None, None
        try:
            content = t_req.body
        except AttributeError:
            pass

        request.content = content
        request.content_length = len(content) if content is not None else None

        if content is not None:
            # parse the POSTed data
            ctype = request.content_type = t_req.headers.get(
                    'CONTENT_TYPE',
                    None,
                    )
            request.form = parse_form(ctype, content)

        # TODO: add more ubiquitous HTTP request headers

        # do session injection in baseclass
        return super(TornadoReflex, self)._do_translate_request(request)

    def _start_response(self, response):
        content = response.content
        status_code = response.status
        enc = response.encoding

        t_req = response.request._native_request

        status_line, headers = gen_http_headers(response)

        # TODO: keep-alive things
        close_found = False
        for k, v in headers:
            if b'connection' == k.lower():
                close_found = True
                break
        if not close_found:
            headers.append((b'Connection', b'close', ))

        header_str = b'\r\n'.join(b'%s: %s' % hdr for hdr in headers)

        # Initiate the actual conversation
        t_req.write(b'HTTP/%s %s\r\n%s\r\n\r\n' % (
                b'1.1' if t_req.supports_http_1_1() else b'1.0',
                status_line, header_str,
                ))

        if response.is_raw_file:
            # push raw file
            return dummy_file_wrapper(response.raw_fp, response.raw_blksz)
        else:
            return send_content_iter(content, enc)

    def _do_deliver_response(self, response):
        if response._vanished:
            return

        req = response.request._native_request
        writer = req.write
        iterable = self._start_response(response)
        for chunk in iterable:
            writer(chunk)
        req.finish()


class WeiyuTornadoAdapter(HTTPServer):
    def __init__(self):
        if HTTPServer is object:
            # Well... the real HTTPServer is not there.
            raise RuntimeError('tornado.httpserver.HTTPServer not found')

        self.__reflex = TornadoReflex()

        # the stimulate method is the callback
        # TODO: keep-alive, chunked encoding, and so on
        super(WeiyuTornadoAdapter, self).__init__(
                self.__reflex.stimulate,
                no_keep_alive=True,
                )


@adapter_hub.register_handler('tornado')
def tornado_adapter_factory(hub):
    return WeiyuTornadoAdapter()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / central registry / classes
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Registry classes
~~~~~~~~~~~~~~~~

This module provides classes for global preference/component registration.
The classes are designed for usage as global singleton instances; put a
registry instance request in your module and register all kinds of things
in it.

'''

from __future__ import unicode_literals, division

import abc
from types import FunctionType, MethodType

import six

# this is to be extended by future declarations
__all__ = ['VALID_REGISTRY_TYPES',
           'FunctionlikeTypes',
           ]

FunctionlikeTypes = (FunctionType, MethodType, )

# enforce some type restrictions
# also to be extended with classes
VALID_REGISTRY_TYPES = {}


def export_registry(cls):
    cls_name = cls.__name__
    __all__.append(cls_name)
    VALID_REGISTRY_TYPES[cls_name] = cls
    return cls


@export_registry
class RegistryBase(six.with_metaclass(abc.ABCMeta)):
    '''A simple registry base class with a ``dict``-like interface.

    The registries are meant for use as global singleton instances, keeping
    record of various components/preferences. The keys and values can both
    be validated and normalized.

    The class is abstract, thus *not* directly usable. You must subclass
    and override its :meth:`normalize_key` and :meth:`normalize_value`
    methods, which are used to implement the said validation/normalization
    behavior.

    '''

    def __init__(self):
        self.__registry = {}

    @abc.abstractmethod
    def normalize_key(self, key):
        '''Called to normalize any key passed in.

        Since the method is abstract, subclasses must override it.

        '''

        pass

    @abc.abstractmethod
    def normalize_value(self, value):
        '''Called to validate and normalize value before storing.

        If ``value`` does not validate (may be because of incorrect types,
        nonsense numerical value, etc), please raise ``ValueError``.

        This method is abstract too, so be sure to override this in your
        subclasses.

        '''

        pass

    def register(self, key, value):
        '''Register a key-value pair into the registry.

        The key and value must both pass validation and normalization, or
        the ``ValueError`` ``raise``'d will be propagated.

        '''

        try:
            normalized_key = self.normalize_key(key)
        except ValueError:
            raise

        if normalized_key in self.__registry:
            raise AttributeError(
                    "'%s' is already registered with value '%s'"
                    % (
                        repr(normalized_key),
                        repr(self.__registry[normalized_key]),
                        )
                    )
        try:
            cooked_value = self.normalize_value(value)
        except ValueError:
            raise

        self.__registry[normalized_key] = cooked_value

    def unregister(self, key):
        '''Remove a previously registered entry.
        '''

        normalized_key = self.normalize_key(key)

        try:
            return self.__registry.pop(normalized_key)
        except KeyError:
            raise KeyError("key '%s' not registered" % repr(normalized_key))

    def __getitem__(self, key):
        return self.__registry[self.normalize_key(key)]

    def __setitem__(self, key, value):
        self.register(key, value)

    def __contains__(self, key):
        return key in self.__registry

    def keys(self):
        return six.iterkeys(self.__registry)

    def values(self):
        return six.itervalues(self.__registry)

    def items(self):
        return six.iteritems(self.__registry)

    def snapshot(self):
        '''Returns a snapshot of current internal state, obtained by calling
        the underlying ``dict`` object's ``copy()`` method.

        '''

        return self.__registry.copy()


@export_registry
class UnicodeRegistry(RegistryBase):
    '''Registry with Unicode keys.

    As its name implies, this class is just :class:`RegistryBase` with keys
    all converted to ``unicode``. Values can still be of any type though.

    '''

    def normalize_key(self, key):
        type_key = type(key)

        if issubclass(type_key, six.text_type):
            return key

        # needs some form of conversion, or worse, decoding...
        if not issubclass(type_key, six.string_types):
            return six.text_type(key)

        # key is a bytestring. ugh. force a UTF-8 encoding
        # Punishment for those under Win32 and an editor not-so-decent,
        # it seems... :-P
        #
        # Unicode{En,De}codeError are both subclass of ``ValueError``,
        # and that's *exactly* what marks a failed validation. Perfect.
        return key.decode('utf-8')

    def normalize_value(self, value):
        return value


@export_registry
class FunctionKeyRegistry(RegistryBase):
    '''Registry with functions as *keys*.

    Mainly for supporting the hooking mechanism, where wrapped functions
    are used as keys to retrieve lists of hooks to execute.

    '''

    def normalize_key(self, key):
        if type(key) not in FunctionlikeTypes:
            raise ValueError("'%s': not a function" % (repr(key), ))

        return key

    def normalize_value(self, value):
        return value


@export_registry
class FunctionValueRegistry(UnicodeRegistry):
    '''Registry for registering functions.

    Note that :meth:`register`'s signature is changed to make
    auto-registration on function definition more Pythonic, which is also
    the registry's desired usage pattern.

    '''

    def normalize_value(self, value):
        if type(value) not in FunctionlikeTypes:
            raise ValueError("'%s': not a function" % (repr(value), ))

        return value

    def register(self, name=None):
        '''Convenient decorator for registering functions.

        If  ``name`` is ``None``, the function's ``__name__`` is used
        as key. Otherwise the name specified is used.

        .. note::
            To use function names, the decorator must be used with parens
            like ``@registry.register()``, or the decorator won't be
            executed.

        '''

        def _decorator_(fn):
            key = fn.__name__ if name is None else name
            super(FunctionRegistry, self).register(key, fn)

            return fn

        return _decorator_


@export_registry
class RegistryRegistry(UnicodeRegistry):
    '''Registry for registering registries.

    In order to ensure true singleton pattern, all registries should be
    acquired from the central registry. Which explains why this very class
    exists...

    See also :mod:`weiyu.registry.provider`, which is the single most
    significant "user" of this class.

    '''

    def normalize_value(self, value):
        if not issubclass(type(value), RegistryBase):
            raise ValueError("'%s': not a registry" % (repr(value), ))

        return value


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / http / WSGI interface
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'WeiyuWSGIAdapter',
            ]

from .. import adapter_hub

from ...helpers.misc import smartstr
from ...registry.provider import request as reg_request
from ...reflex.classes import ReflexRequest

from .base import BaseHTTPReflex
from .util import status_to_str, dummy_file_wrapper, send_content_iter
from .util import parse_form, gen_http_headers


class WSGIRequest(ReflexRequest):
    def __init__(self, env, start_response, site_conf):
        super(WSGIRequest, self).__init__(env)

        self.start_response = start_response
        self.site = site_conf


class WSGIReflex(BaseHTTPReflex):
    def __init__(self):
        super(WSGIReflex, self).__init__()

    def _do_accept_request(self, env, start_response):
        return WSGIRequest(env, start_response, self.SITE_CONF)

    def _do_translate_request(self, request):
        # populate some useful fields using WSGI env
        env = request.env
        # decode the path bytestring
        # TODO: improve encoding handling here
        path = request.path = smartstr(env['PATH_INFO'], 'utf-8', 'replace')

        # Move routing (much) earlier so we don't waste time in processing
        # requests impossible to fulfill.
        # Note that we don't pass in "request" at this moment. The object
        # can be replaced by potential hooks, and we certainly don't want
        # a reference to be frozen in the request.
        # Return value is of format (fn, args, kwargs, route_data, )
        route_result = self._do_routing(path)
        request.callback_info = route_result[:-1]
        request.route_data = route_result[-1]

        # Rest of request object preparation goes here...
        request.remote_addr = smartstr(env['REMOTE_ADDR'])
        method = request.method = smartstr(env['REQUEST_METHOD'])
        length, _env_length = None, None
        try:
            _env_length = smartstr(env['CONTENT_LENGTH'], 'ascii', 'ignore')
        except KeyError:
            pass

        if _env_length is not None and len(_env_length) > 0:
            try:
                length = int(_env_length)
            except ValueError:
                pass
        request.content_length = length

        # read the request body if content length is given
        if length is not None and request.method == 'POST':
            request.content = env['wsgi.input'].read(length)

            # parse the POSTed data
            ctype = request.content_type = env.get('CONTENT_TYPE', None)
            request.form = parse_form(ctype, request.content)
        else:
            request.content = None

        # TODO: add more ubiquitous HTTP request headers

        # do session injection in baseclass
        return super(WSGIReflex, self)._do_translate_request(request)

    def _do_deliver_response(self, response):
        if response._vanished:
            # We've been hijacked. Nothing should be done.
            return

        content = response.content
        status_code = response.status
        enc = response.encoding

        request = response.request

        # Initiate the actual conversation
        request.start_response(*gen_http_headers(response))

        if response.is_raw_file:
            # push raw file using file_wrapper if provided
            # if one is absent, use a dummy one instead
            raw_fp, raw_blksz = response.raw_fp, response.raw_blksz

            if 'wsgi.file_wrapper' in request.env:
                file_wrapper = request.env['wsgi.file_wrapper']
            else:
                file_wrapper = dummy_file_wrapper

            if raw_blksz is None:
                return file_wrapper(raw_fp)
            else:
                return file_wrapper(raw_fp, raw_blksz)
        else:
            return send_content_iter(content, enc)


class WeiyuWSGIAdapter(object):
    def __init__(self):
        self.reflex = WSGIReflex()

    def __call__(self, env, start_response):
        return self.reflex.stimulate(env, start_response)


@adapter_hub.register_handler('wsgi')
def wsgi_adapter_factory(hub):
    return WeiyuWSGIAdapter()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / helpers / metadata annotation
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Object annotation
~~~~~~~~~~~~~~~~~

This module provides managed storage of arbitrary annotations on Python
objects (mainly functions).

The annotations are stored in a ``dict`` injected into the object's
``__dict__`` by the key ``'_weiyu_data_'``. You can of course operate on
the dict yourself, but by using this module's code paths compatibility is
realized in case of renaming of the annotation dictionary.

Functions
---------

'''

from __future__ import unicode_literals, division

__all__ = [
        'is_annotated',
        'ensure_annotation',
        'annotate',
        'get_annotation',
        'remove_annotations',
        ]

ANNOTATE_KEY = '_weiyu_data_'


def is_annotated(obj):
    '''Returns if an object has been annotated by this module;
    that is, have a ``_weiyu_data_`` dict implanted into its ``__dict__``.

    '''

    return ANNOTATE_KEY in obj.__dict__


def ensure_annotation(obj):
    '''Ensures the presence of ``_weiyu_data_`` dict inside the object's
    ``__dict__``. Does nothing if the object has already been annotated
    somehow.

    '''

    if not is_annotated(obj):
        obj.__dict__[ANNOTATE_KEY] = {}


def annotate(obj, **kwargs):
    '''Annotates ``obj`` with keyword arguments supplied.

    This is implemented by calling the underlying ``dict``\ 's ``update``
    method.

    '''

    ensure_annotation(obj)
    obj.__dict__[ANNOTATE_KEY].update(kwargs)


def remove_annotations(obj, keys):
    '''Removes annotation keys ``keys`` from ``obj``.

    Raises :exc:`AttributeError` if ``obj`` is considered not previously
    annotated by :func:`.is_annotated`. Keys that do not exist in the
    annotation dictionary are silently ignored.

    .. note::
        This function does not purge the annotation dictionary from the
        object's ``__dict__``, even if that dict becomes empty after
        removal.

    '''

    if not is_annotated(obj):
        raise AttributeError(
        'object %s is not annotated' % (
            repr(obj),
            )
        )

    data = obj.__dict__[ANNOTATE_KEY]
    for key in keys:
        if key in data:
            del data[key]


def get_annotation(obj, key=None):
    '''Retrieves annotation from ``obj``.

    Raises :exc:`AttributeError` if ``obj`` is not previously annotated
    by this module.

    If ``key`` is ``None``, returns a copy of the object's annotation
    dictionary; returns the data associated with ``key`` otherwise. If
    ``key`` is not found, :exc:`KeyError` is ``raise``\ d.

    '''

    if not is_annotated(obj):
        raise AttributeError(
                'object %s is not annotated' % (
                    repr(obj),
                    )
                )

    data = obj.__dict__[ANNOTATE_KEY]
    return data.copy() if key is None else data[key]


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / dummy renderer
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Dummy renderer
--------------

This renderer, as its name suggests, does not render anything. This is
useful when implementing certain API endpoints that do not or must not
return any content.

'''

from __future__ import unicode_literals, division

__all__ = ['DummyRenderable', ]

from . import render_hub
from .base import Renderable


class DummyRenderable(Renderable):
    def _do_render(self, result, context):
        # Intentionally left blank
        return b'', {}


@render_hub.register_handler('dummy')
def dummy_render_handler(hub, name, *args, **kwargs):
    # template name is meaningless, thus ignored.
    return DummyRenderable(*args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / db drivers / pymongo driver
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
``pymongo`` Database Driver
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This is the MongoDB driver written for ``weiyu``, using ``pymongo``.

'''

from __future__ import unicode_literals, division

from functools import wraps

import pymongo

from .. import db_hub
from .baseclass import BaseDriver


class PymongoDriver(BaseDriver):
    '''``pymongo`` driver class.'''

    def __init__(
            self,
            path,
            host,
            port,
            is_replica,
            max_pool_size,
            tz_aware,
            ):
        '''Constructor function.

        The database server is specified through the parameters ``host`` and
        ``port``; individual database is selected by ``path``.

        If the database to connect to is actually a replica set, set
        ``is_replica`` to ``True``. This will make the connection a
        ``MongoReplicaSetClient`` instead of ``MongoClient``.

        Other parameters have the same meanings as in PyMongo.

        '''

        super(PymongoDriver, self).__init__()

        self.host, self.port, self.path = host, port, path
        self.max_pool_size, self.tz_aware = max_pool_size, tz_aware

        _conn_type = (pymongo.MongoReplicaSetClient
                           if is_replica
                           else pymongo.MongoClient
                           )
        self.conn = _conn_type(
                host=host,
                port=port,
                max_pool_size=max_pool_size,
                tz_aware=tz_aware,
                auto_start_request=False,
                )
        self.db = self.conn[self.path]
        self._buckets = {}

    def start(self):
        self.conn.start_request()

    def finish(self):
        self.conn.end_request()

    def get_bucket(self, bucket):
        try:
            return self._buckets[bucket]
        except KeyError:
            b = getattr(self.db, bucket)
            self._buckets[bucket] = b

        return self._buckets[bucket]

    def __repr__(self):
        return b'<weiyu.db/pymongo: %s:%d, path=%s>' % (
                self.host,
                self.port,
                self.path,
                )


@db_hub.register_handler('pymongo')
def pymongo_handler(
        hub,
        path,
        host='127.0.0.1',
        port=27017,
        is_replica=False,
        max_pool_size=10,
        tz_aware=False,
        ):
    return PymongoDriver(
            path,
            host,
            port,
            is_replica,
            max_pool_size,
            tz_aware,
            )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / top-level package / package version file
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# NOTE: the SVN support code reused Django project's.

from __future__ import unicode_literals, division

__all__ = ['VERSION_MAJOR', 'VERSION_MINOR', 'VERSION_REV',
           'VERSION', 'VERSION_DEV', 'VERSION_STR',
           ]

import os.path
from sys import getfilesystemencoding
import re
import weiyu as __this_pkg

# Caveat: this file is imported also by setup.py, so we can't guarantee
# six's presence! This can lead package installation to fail.
# import six
#
# Instead, make up ourselves the only feature we need here -- six.text_type,
# which is very easy to implement...
text_type = str if __import__('sys').version_info[0] == 3 else unicode


# Version information.
VERSION_MAJOR = 0
VERSION_MINOR = 1
VERSION_REV = 1

VERSION = (VERSION_MAJOR, VERSION_MINOR, VERSION_REV, 'final', 0)

_VCS_HANDLERS = []


def get_version():
    version = '%s.%s' % (VERSION[0], VERSION[1])
    if VERSION[2]:
        version = '%s.%s' % (version, VERSION[2])
    if VERSION[3:] == ('alpha', 0):
        version = '%s-dev' % version
    else:
        if VERSION[3] != 'final':
            version = '%s %s %s' % (version, VERSION[3], VERSION[4])

    return version


###############################################################
## for getting revision info from a VCS
###############################################################


def get_vcs_revision(path=None):
    # try the handlers one by one, return the first successful
    # match of VCS info
    # cast path into unicode if necessary
    if path is not None and not isinstance(path, text_type):
        path = path.decode(getfilesystemencoding())

    for handler in _VCS_HANDLERS:
        is_ok, result = handler(path)
        if is_ok:
            return (is_ok, result, )

    # no info available, signal failure
    return (False, None, )


def get_svn_revision(path=None):
    '''Returns the SVN revision in the form SVN-XXXX, where XXXX is the
    revision number.

    Returns (False, None) if anything goes wrong, such as an unexpected
    format of internal SVN files.

    If path is provided, it should be a directory whose SVN info you want to
    inspect. If it's not provided, this will use the root package
    directory.

    '''

    rev = None
    if path is None:
        path = __this_pkg.__path__[0]
    entries_path = '%s/.svn/entries' % path

    try:
        with open(entries_path, 'r') as fp:
            entries = fp.read().decode('ascii', 'replace')
    except IOError:
        return (False, None, )
    else:
        # Versions >= 7 of the entries file are flat text. The first line
        # is the version number. The next set of digits after 'dir' is the
        # revision.
        if re.match(r'(\d+)', entries):
            rev_match = re.search(r'\d+\s+dir\s+(\d+)', entries)
            if rev_match:
                rev = rev_match.groups()[0]
        # Older XML versions of the file specify revision as an attribute
        # of the first entries node.
        else:
            from xml.dom import minidom
            dom = minidom.parse(entries_path)
            rev = (dom.getElementsByTagName('entry')[0]
                    .getAttribute('revision')
                    )

    if rev:
        return (True, 'SVN-%s' % rev, )
    return (False, None, )

_VCS_HANDLERS.append(get_svn_revision)


def get_git_commit(path=None):
    '''Returns the Git commit id in the form Git-xxxxxxxx,
    where xxxxxxxx is the commit hash truncated after the 8th hex digit.

    Returns (False, None) if anything goes wrong.

    If path is provided, it should be a directory whose Git info you want
    to inspect. If it's not provided, this will use the root package
    directory's parent dir.

    '''

    # Git directory...
    if path is None:
        path = __this_pkg.__path__[0]
        git_path = '%s/../.git/' % path
    else:
        git_path = '%s/.git/' % path

    # normalize a little bit
    git_path = os.path.normpath(git_path)

    # Phase 1, read the HEAD file to get the current head
    head_path = os.path.join(git_path, 'HEAD')
    try:
        with open(head_path, 'rb') as fp:
            ref = fp.read().strip().decode('ascii', 'replace')
    except IOError:
        return (False, None, )
    else:
        ref_path_match = re.search(r'^ref: (.*)$', ref)
        if ref_path_match:
            ref_path = ref_path_match.groups()[0]
        else:
            # unrecognized HEAD format...
            return (False, None, )

    # now ref_path is ready, move on to Phase 2, pull out the commit id
    commit_path = os.path.normpath(os.path.join(git_path, ref_path))
    try:
        with open(commit_path, 'rb') as fp:
            commit = fp.read().strip().decode('ascii', 'replace')
    except IOError:
        return (False, None, )
    else:
        # Truncate the commit id.
        commit_id = commit[:8]

    # if we arrive here, we're done and commit id is ready.
    return (True, 'Git-%s' % commit_id, )

_VCS_HANDLERS.append(get_git_commit)

# init our version strings... they are constant during one run
_verstr = get_version()

# VCS revision, if we are in an devel environment
# updated to support Git too
_vcs_ok, _vcs_rev = get_vcs_revision()

VERSION_DEV = _vcs_rev if _vcs_ok else ''
VERSION_STR = _verstr if not _vcs_ok else ' '.join([_verstr, _vcs_rev])

del _vcs_ok, _vcs_rev, _verstr


# vi:ai:et:ts=4 sw=4 sts=4 fenc=utf-8



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / base functionality
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'RouterTargetBase',
            'RouterBase',
            'STATUS_REACHED',
            'STATUS_FORWARD',
            'STATUS_NOROUTE',
            ]

from collections import OrderedDict

import six

from ..helpers.misc import smartstr

STATUS_REACHED, STATUS_FORWARD, STATUS_NOROUTE = range(3)


def is_router(tgt):
    return isinstance(tgt, (RouterBase, ))


class DispatchError(Exception):
    pass


# XXX Make up clearer names for all those "target"s!

class RouterBase(object):
    def __init__(
            self,
            target_initializers,
            name=None,
            parent=None,
            scope='',
            ):
        self.name, self.scope, self.parent = name, scope, parent
        self.route_table = [
                self.__class__.target_type(router=self, *i)
                for i in target_initializers
                ]

        # Fix sub-routers' parents.
        for tgt in self.route_table:
            if tgt.target_is_router:
                tgt.target.parent = tgt

    def lookup(self, querystr, prev_args=None, prev_kwargs=None):
        # XXX is this needed, or am I overly sensitive?
        querystr = smartstr(querystr)

        if prev_args is None:
            prev_args = []

        if prev_kwargs is None:
            prev_kwargs = {}

        # sequentially check our "route table"
        for entry in self.route_table:
            # return is of form ``(hit, target, args, kwargs, )``
            result = entry.lookup(
                    querystr,
                    prev_args,
                    prev_kwargs,
                    )

            if result[0]:
                # successful match
                return result

        # match failure
        return (False, None, None, None, None, )

    def dry_dispatch(self, querystr, *args):
        '''Do all things except actually invoking callback function,
        returns the calculated parameters that can be used to do a real
        dispatch.

        '''

        hit, target, more_args, kwargs, data = self.lookup(querystr)

        if not hit:
            raise DispatchError(
                    "query string %s: nowhere to dispatch" % (
                        repr(querystr),
                        ))

        # append resolved positional args to args passed in
        extended_args = list(args)
        extended_args.extend(more_args)

        return target, extended_args, kwargs, data

    def dispatch(self, querystr, *args):
        target, ext_args, kwargs, data = self.dry_dispatch(querystr, *args)
        return target(*ext_args, **kwargs)

    def build_reverse_map(self):
        scope = self.scope
        map_ = {scope: {}}  # OrderedDict()
        this_scope = map_[scope]

        # According to the ctor, route_table is guaranteed to be a list,
        # so reversing using slicing syntax is fine.
        for entry in self.route_table[::-1]:
            if entry.target_is_router:
                chld_map = entry.target.reverse_map
                for chld_scope, chld_scopemap in six.iteritems(chld_map):
                    if chld_scope in map_:
                        map_[chld_scope].update(chld_scopemap)
                    else:
                        map_[chld_scope] = chld_scopemap
            else:
                tgt_spec = entry.data.get('target_spec', None)
                this_scope[tgt_spec] = entry.reverse_pattern

        return map_

    @property
    def reverse_map(self):
        try:
            return self._reverse_map
        except AttributeError:
            map_ = self._reverse_map = self.build_reverse_map()
            return map_


class RouterTargetBase(object):
    def __init__(self, target, extra_data=None, router=None):
        self.target, self.data, self.router = target, extra_data, router

        # for nested processing
        self.target_is_router = is_router(target)

    def check(self, querystr):
        # return is assumed to be in the form of
        # (hit_status, args, kwargs, new_qs, )
        # since it is unnecessary and redundant to return self.target
        # to an internal method
        raise NotImplementedError

    def lookup(self, querystr, prev_args, prev_kwargs):
        # return value is of form (hit, target, args, kwargs, )
        # result is of form (status, args, kwargs, new_querystr, )
        status, args, kwargs, new_qs = self.check(
                querystr,
                prev_args,
                prev_kwargs,
                )

        if status == STATUS_REACHED:
            # reached end, make args a tuple
            # XXX Is this necessary?
            return (True, self.target, tuple(args), kwargs, self.data, )
        elif status == STATUS_NOROUTE:
            return (False, None, None, None, None, )
        elif status == STATUS_FORWARD:
            # do nested routing
            return self.target.lookup(new_qs, args, kwargs)

        # unreachable if check function behaves properly
        raise DispatchError('Impossible check result %s, bug detected'
                % repr((status, args, kwargs, new_qs, ))
                )

    @property
    def reverse_pattern(self):
        try:
            return self._reverse_pattern
        except AttributeError:
            pat = self._get_reverse_pattern()
            if self.router.parent is not None:
                parent_pat = self.router.parent.reverse_pattern
            else:
                # no parent target, use an empty one
                parent_pat = ('', set(), )

            # merge the parent's reverse pattern
            pat_str = parent_pat[0] + pat[0]
            pat_vars = set(parent_pat[1])
            pat_vars.update(pat[1])
            self._reverse_pattern = (pat_str, pat_vars, )
            return self._reverse_pattern

    def _get_reverse_pattern(self):
        return


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / central registry / config loader
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Configuration loader
~~~~~~~~~~~~~~~~~~~~

As much of weiyu's functionality relies on the registries, there should be
a convenient way of pre-loading configuration data before performing any
significant actions. This module serves just that purpose, providing several
configuration formats free to choose from; of course, you are also free to
roll your own loader by inheriting :class:`.BaseConfig` and implement the
``_do_loads`` method. At present, only JSON and Python pickles are
available for use; more is expected to be added.


Basics
------

Global configuration files which is used by
:meth:`BaseConfig.populate_central_regs` to populate (possibly) multiple
registries must have a Pythonized format like this::

    {
        u'registry.name.1': {
            u'class': u'ClassNameOfRegistry', # must be bytes or unicode
            u'no_update': True or False,
            u'data': {
                # this is the actual configuration data
                u'spam': u'foo',
                u'eggs': 42,
                u'etc': u'etc',
                }
            },

        # optionally more registries, or directives to follow
    }

The ``class`` attribute must specify one of the registry types "exported"
in :mod:`weiyu.registry.classes`; this limitation may be lifted in future
versions of weiyu.

The ``no_update`` boolean, whose value is directly passed as the
``nodup`` parameter to :func:`weiyu.registry.provider.request`,
indicates that an exception should be raised when a registry with the
same name already exists (possibly initialized to an empty one by its
associated module). It really should be set to ``False`` for populating
registries initialized by modules. Only set this to ``True`` when you are
sure that no module will try to create the registry before you load the
configuration.


Preprocessing directives
------------------------

Directives are just sections in the config file with names starting with
``$$``. They are then parsed by :meth:`BaseConfig.process_directives`.

Currently there is only one directive implemented: the (top-level) include
directive ``$$include``. Its parameter is a ``list`` of files to source:
content of the files are read and processed in order, overwriting values
if they already exist. It does not support including files into a variable,
for example, at the moment. Also it does not detect circular includes, any
such situation the loader will loop infinitely and eventually crash with a
:exc:`MemoryError`. Don't try that.


Classes
-------

'''

from __future__ import unicode_literals, division

__all__ = [
        'BaseConfig',
        'JSONConfig',
        'PickleConfig',
        ]

from os.path import abspath, splitext
from functools import wraps
import abc
import json

import six
from six.moves import cPickle as pickle

from .classes import VALID_REGISTRY_TYPES
from .provider import request

# $$directives
DIRECTIVE_CLASS_OVERRIDE = '$$class'
DIRECTIVE_UPDATE_OVERRIDE = '$$nodup'
DIRECTIVE_INCLUDE = '$$include'

# config file extension handler registry
_EXTENSION_HANDLERS = {}


def handler_for_ext(ext):
    def _decorator_(thing):
        _EXTENSION_HANDLERS[ext] = thing
        return thing
    return _decorator_


class BaseConfig(six.with_metaclass(abc.ABCMeta)):
    '''Base class for prepopulated config storages. It has an interface
    inspired by ``pickle``: methods such as :meth:`loads` and :meth:`load`
    exist and more or less behaves like the standard library counterpart.

    The :meth:`_do_loads` method is abstract, it is up to you to decide the
    way configuration files map to Python object in a subclass.

    '''

    @staticmethod
    def get_config(path):
        ext = splitext(path)[1]

        try:
            cls = _EXTENSION_HANDLERS[ext]
        except KeyError:
            raise ValueError(
                    "cannot handle config file '%s' with extension '%s'" % (
                        path,
                        ext,
                        ),
                    )

        return cls(path)

    def __init__(self, path=None):
        '''Constructor function.

        :param path: Path to config file. Defaults to ``None``; If this is
        the case, data may be loaded by directly calling :meth:`loads` or
        :meth:`load` with the appropriate type of data source.

        '''

        self.path, self.data = path, None

    @abc.abstractmethod
    def _do_loads(self, s, *args, **kwargs):
        '''Constructs a Python object from input. This method is abstract.

        :param s: The input string.

        .. note::

            Don't do preprocessing in this method; they are to be done by
            :meth:`loads` and the other interface methods. Just construct
            the "raw" deserialized form.

        '''

        pass

    def loads(self, s, *args, **kwargs):
        '''Loads configuration from a string.
        '''

        raw_data = self._do_loads(s, *args, **kwargs)
        self.data = self.process_directives(raw_data)
        return self.data

    def load(self, fp=None, *args, **kwargs):
        '''Loads configuration from a file-like object with a ``read`` method.

        .. note::

            The implementation at this point is naive. It just tries to read
            the whole input, then feeds the string into :meth:`loads`. Beware
            if the input is potentially very large.

        '''

        # XXX this function's a naive wrapper to get rid of some duplication

        if fp is None:
            # invoke ``self.load_from_path`` to read the file specified
            # by ``self.path``
            return self.load_from_path(None)

        # Unicode cleanness
        return self.loads(
                fp.read().decode('utf-8', 'replace'),
                *args, **kwargs
                )

    def load_from_path(self, path=None):
        '''Loads configuration from path ``path``; if ``path`` is ``None``,
        try to use ``self.path``. If that value is also ``None``, a
        :exc:`ValueError` is ``raise``\ d.

        '''

        if path is None:
            if self.path is None:
                raise ValueError('no usable path configured')
            path = self.path

        return self._do_load_from_path(path)

    def _do_load_from_path(self, path):
        with open(path, 'rb') as fp:
            return self.load(fp)

    def do_include(self, includes):
        tmp = {}
        for path in includes:
            # XXX FIXME: Infinite includes is possible!!
            # NOTE Security is important here, so paths should be
            # at least canonicalized.
            real_path = abspath(path)
            tmp.update(self.load_from_path(real_path))
        return tmp

    def process_directives(self, data):
        # process '$$include': [inc1, inc2, etc, ]
        if DIRECTIVE_INCLUDE in data:
            include_list = data.pop(DIRECTIVE_INCLUDE)
            data.update(self.do_include(include_list))

        return data

    def ensure_data_presence(self):
        '''When called, checks if ``self.data`` is ``None``. If it is the
        case, calls :meth:`load_from_path` to load configuration from
        ``self.path``. Note that this method does NOT ensure the
        *consistency* between the stored state and the config file's data;
        it just checks for presence.

        '''

        if self.data is None:
            self.load_from_path(None)

    def populate_registry(self, registry):
        '''Populates a single registry instance. For filling the global
        registry, use :meth:`populate_central_regs`.

        '''

        self.ensure_data_presence()
        for k, v in six.iteritems(self.data):
            registry.register(k, v)

    def populate_central_regs(self):
        '''Fills the central registry, i.e. the
        :class:`.classes.RegistryRegistry` singleton. Config data format
        used for this method is described above.

        '''

        self.ensure_data_presence()
        for k, v in six.iteritems(self.data):
            # k is registry name, v is a dict
            # v's format is as follows:
            # {['$$class': '<class name of registry (optional)>',]
            #  ['$$nodup': True/False (optional),]
            #  key1: val1,
            #  key2: val2,
            #  etc: etc,
            #  }
            # first a little type sanity check
            if not isinstance(k, six.string_types):
                raise ValueError(
                        'registry name must be string; got %s (of type %s)'
                        % (repr(k), repr(type(k)), )
                        )

            if not isinstance(v, dict):
                raise ValueError(
                        'pythonized config entry must be dict; got type %s'
                        % (repr(type(v)), )
                        )

            # extract class and updating behavior
            if DIRECTIVE_CLASS_OVERRIDE in v:
                class_name = v.pop(DIRECTIVE_CLASS_OVERRIDE)
            else:
                # hard-code to be a string-keyed registry
                class_name = 'UnicodeRegistry'

            if DIRECTIVE_UPDATE_OVERRIDE in v:
                nodup = v.pop(DIRECTIVE_UPDATE_OVERRIDE)
            else:
                # allow updating existing registries through conf by default
                nodup = False

            # override field type check
            if not isinstance(class_name, six.string_types):
                raise ValueError('registry class name not of string type')

            if not isinstance(nodup, bool):
                raise ValueError('no_update property must be boolean')

            # class validation
            try:
                cls = VALID_REGISTRY_TYPES[class_name]
            except KeyError:
                raise ValueError("registry class name '%s' not valid"
                        % (class_name, )
                        )

            # format check (finally...) passed, go on with actually updating
            # NOTE: obviously registries cannot have duplicate names,
            # here we'll rely on RegistryBase's behavior to ensure this :P
            reg = request(k, autocreate=True, nodup=nodup, klass=cls)

            # populate the dict
            # NOTE: We don't use update() here, because filling values one by
            # one would trigger validation and normalization for each of the
            # k-v pairs, and sanity is what we need.
            for reg_k, reg_v in six.iteritems(v):
                reg[reg_k] = reg_v


@handler_for_ext('.json')
class JSONConfig(BaseConfig):
    '''JSON config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        dump = json.loads(s, *args, **kwargs)
        # TODO: some canonicalization or such
        return dump


@handler_for_ext('.pickle')
class PickleConfig(BaseConfig):
    '''Python pickle config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        return pickle.loads(s, *args, **kwargs)


@handler_for_ext('.yml')
@handler_for_ext('.yaml')
class YAMLConfig(BaseConfig):
    '''YAML config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        import yaml
        from yaml import Loader, SafeLoader

        # Force Unicode string output according to this SO question
        # 2890146/how-to-force-pyyaml-to-load-strings-as-unicode-objects
        def construct_yaml_str(self, node):
            # Override the default string handling function
            # to always return unicode objects
            return self.construct_scalar(node)

        _STR_TAG = 'tag:yaml.org,2002:str'
        Loader.add_constructor(_STR_TAG, construct_yaml_str)
        SafeLoader.add_constructor(_STR_TAG, construct_yaml_str)

        # TODO: optionally utilize C acceleration if available
        return yaml.load(s, *args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / http / base class
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'BaseHTTPReflex',
            ]

from ...helpers.misc import smartbytes

from ...registry.provider import request as reg_request

from ...reflex.classes import BaseReflex
from ...router import router_hub
from ...session import session_hub
from ...signals import signal_hub
from ...rendering import render_hub

from .. import adapter_hub

# Status codes that cannot have response body
# Used to prevent rendering code from being invoked
# TODO: add more of them
NO_RESP_BODY_STATUSES = {405, }


@adapter_hub.declare_middleware('session')
class HTTPSessionMiddleware(object):
    def __init__(self):
        # get options
        site = reg_request('site')

        # assume session config is present if this middleware is used
        ses_conf = site['session']
        ses_backend, ses_opts = ses_conf['backend'], ses_conf['options']

        # request a session backend
        self._backend = session_hub.do_handling(ses_backend, ses_opts)

    def do_pre(self, request):
        self._backend.preprocess(request)

    def do_post(self, response):
        self._backend.postprocess(response)


class BaseHTTPReflex(BaseReflex):
    def __init__(self):
        self.SITE_CONF = reg_request('site')
        # no more repeated lookups for handler
        # NOTE: Remember this can lead to stale value when handler got
        # dynamically updated. This is (admittedly) a VERY rare and
        # dangerous case that's simply NOT going to work, but the expected
        # behavior is documented here anyway.
        self._do_routing = router_hub.get_handler('http')

    def _do_translate_request(self, request):
        # Middleware
        # TODO: ability to skip response generation?
        signal_hub.fire_nullok('http-middleware-pre', request)

        return request

    def _do_generate_response(self, request):
        fn, args, kwargs = request.callback_info
        response = fn(request, *args, **kwargs)

        # Middleware
        signal_hub.fire_nullok('http-middleware-post', response)

        return response

    def _do_postprocess(self, response):
        # Render the response early
        # TODO: find a way to extract all the literals out
        request = response.request
        ctx, hdrs, extras = response.context, [], {}
        render_in = cont = mime = None
        dont_render = False

        # are we hijacked by some other library (e.g. socketio)?
        if ctx.get('request_vanished', False):
            # All further actions starting from accepting request are
            # performed by the other library, so we can't do anything
            # in this case. Just bail out.
            response._vanished = True
            return response

        response._vanished = False

        # encoding
        enc = response.encoding = ctx.get('enc', 'utf-8')

        # is this a raw file push request?
        response.is_raw_file = ctx.get('is_raw_file', False)
        if response.is_raw_file and 'sendfile_fp' in response.content:
            # request to send raw file valid, suppress rendering
            dont_render = True
            response.raw_fp = response.content['sendfile_fp']
            response.raw_blksz = response.content.get('blocksize', None)

            # If rendering needs to be suppressed and content is not a raw
            # file, we cannot safely assume the data stream is of type
            # text/html then...
            mime = ctx.get('mimetype', 'application/octet-stream')

        # is this an error response which cannot have response body?
        # TODO: add more status codes
        if response.status in NO_RESP_BODY_STATUSES:
            dont_render = True

        if not dont_render:
            if issubclass(type(request.route_data), dict):
                # mapping object, see if we could get the hint...
                render_in = request.route_data.get('render_in', None)

            if render_in is None:
                raise TypeError(
                        "Rendering is not suppressed, but don't know "
                        "where to get rendering instruction!"
                        )

            # rendering is not suppressed, do it now
            cont, extras = render_hub.render_view(
                    request.callback_info[0],
                    response.content,
                    ctx,
                    render_in,
                    )

            if 'mimetype' in extras:
                # Allow renderers to override mimetype
                mime = extras['mimetype']
            else:
                if mime is None:
                    mime = ctx.get('mimetype', 'text/html')

            # encode content, if it's a Unicode thing
            response.content = smartbytes(cont, enc, 'replace')

        # TODO: convert more context into HTTP headers as much as possible
        # generate Content-Type from mimetype and charset
        # but don't append charset if the response is a raw file
        if response.is_raw_file:
            contenttype = mime
        else:
            contenttype = '%s; charset=%s' % (mime, enc, )

        if not dont_render:
            hdrs.append((b'Content-Type', contenttype, ))

        # generate Set-Cookie from cookies
        for cookie_line in ctx.get('cookies', []):
            hdrs.append((b'Set-Cookie', cookie_line, ))

        # 405 Not Allowed header
        if response.status == 405:
            hdrs.append((b'Allow', ctx['allowed_methods'], ))

        response.http_headers = hdrs
        response._dont_render = dont_render

        return response


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / routing config / lexer
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division, print_function

import re
from collections import OrderedDict

import six


class Token(object):
    def __init__(self, type, data, span, line):
        self.type = type
        self.value = data
        self.span = span
        self.line = line

    def __repr__(self):
        return '%d, %d-%d Token(%s, %s)' % (
                self.line,
                self.span[0],
                self.span[1],
                self.type,
                repr(self.value),
                )


class Lexer(object):
    def __init__(self, s=None):
        self._lex_reset(s)

    def _lex_reset(self, s):
        self._pos = 0
        self._lineno = 1

        if s is not None:
            self._input = s
            self._length = len(s)

        self._tokens = None
        self._state = 'INITIAL'
        self._prev = None
        self._ply_pos = None

    def input(self, s):
        self._lex_reset(s)

    def get_token(self, type, frag, match):
        span = match.span()
        return Token(type, frag, span, self._lineno)

    def eof(self):
        return self._pos >= self._length

    def lex(self):
        cls = self.__class__

        _rule_cache = OrderedDict(
                (name, re.compile(rule), )
                for name, rule in six.iteritems(cls.RULES)
                )
        _handlers = {
                name: getattr(self, 't_%s' % name)
                for name in six.iterkeys(_rule_cache)
                }

        self._tokens = []
        while self._pos <= self._length:
            matched = False
            for name, rule in six.iteritems(_rule_cache):
                match = rule.match(self._input, self._pos)
                if match is not None:
                    matched = True
                    match_span = match.span()
                    # advance the pointer
                    self._pos = match_span[1]

                    # generate the token(s)
                    frag = self._input[match_span[0]:match_span[1]]
                    skip, result = _handlers[name](frag, self, match)
                    if skip:
                        continue

                    # print result
                    self._tokens.extend(result)

                    # update previous token
                    self._prev = self._tokens[-1].type
                    break

            if not matched:
                raise ValueError('could not advance any more')

    def token(self):
        # For compatibility with PLY.
        if self._ply_pos is None:
            self._ply_pos = 0
            if self._tokens is None:
                # lex first
                self.lex()
        if self._ply_pos >= len(self._tokens):
            return None

        pos = self._ply_pos
        self._ply_pos += 1
        return self._tokens[pos]


class WRLexer(Lexer):
    RULES = OrderedDict([
            ('ATTRIB', r'--[^\s]+'),
            ('NEWLINE', r'\n+'),
            ('SPACE', r'[ \t]*'),
            ('COLON', r':\n'),
            #('LITERAL', r'(?:[^\s]*\\\s)*[^\s]+'),
            ('LITERAL', r'[^\'\"\s]*[^\'\"\s:]|\'[^\']*\'|\"[^\"]*\"'),
            ('EOF', r'$'),
            ])

    tokens = list(six.iterkeys(RULES)) + ['INDENT', 'DEDENT', ]
    tokens.remove('SPACE')
    tokens = tuple(tokens)

    def __init__(self, *args, **kwargs):
        super(WRLexer, self).__init__(*args, **kwargs)
        self._is_prev_space_empty = False
        self._indent_stack = [0]

    # f for fragment, l for lexer instance, m for match object
    def t_ATTRIB(self, f, l, m):
        return False, [self.get_token('ATTRIB', f, m), ]

    def t_NEWLINE(self, f, l, m):
        l._lineno += len(f)
        return False, [self.get_token('NEWLINE', f, m), ]

    def t_LITERAL(self, f, l, m):
        #return False, [self.get_token('LITERAL', f, m), ]
        actual_frag = f if f[0] not in '\'"' else f[1:-1]
        return False, [self.get_token('LITERAL', actual_frag, m), ]

    def t_SPACE(self, f, l, m):
        # but we certainly don't want to be stuck in an infinite loop, so
        # record the fact if we ARE matching an empty string right now
        if self._is_prev_space_empty:
            self._is_prev_space_empty = False
            return True, None

        if len(f) == 0:
            self._is_prev_space_empty = True

        if not l.eof() and l._input[m.span()[0] - 1] != '\n':
            # we're only interested in leading whitespaces
            # INCLUDING EOF CONDITION
            return False, []

        # try to generate INDENT/DEDENT tokens
        # first let's calculate effective width
        efflen = 0
        for ch in f:
            efflen += (8 - efflen % 8) if ch == '\t' else 1

        indstk = self._indent_stack
        stktop = indstk[-1]

        # same indentation level? if is, return early
        if stktop == efflen:
            return False, []

        # INDENT handling
        if stktop < efflen:
            # indent, push stack
            indstk.append(efflen)
            return False, [self.get_token('INDENT', f, m), ]

        # DEDENT check
        dedents = []
        while stktop > efflen:
            dedents.append(self.get_token('DEDENT', f, m), )
            # pop stack
            indstk.pop()
            stktop = indstk[-1]
        if stktop != efflen:
            # inconsistent indentation, bark
            raise IndentationError('inconsistent indentation')
        # check ok, return
        return False, dedents

    def t_COLON(self, f, l, m):
        return False, [self.get_token('COLON', f, m), ]

    def t_EOF(self, f, l, m):
        # just treat this as whitespace...
        result = self.t_SPACE(f, l, m)
        # but we must make sure the lexing loop terminates!
        l._pos += 1
        return result


if __name__ == b'__main__':
    import sys
    fname = sys.argv[1]
    with open(fname, 'rb') as fp:
        content = fp.read().decode('utf-8')

    m = WRLexer()
    m.input(content)
    while True:
        tok = m.token()
        if tok is None:
            break
        print(tok)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / central registry / service provider singleton
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Registry service provider
~~~~~~~~~~~~~~~~~~~~~~~~~

This module is the "official" provider of registry instances, in which a
:class:`.classes.RegistryRegistry` singleton lives.

To ensure true singleton-ness, all registries should be acquired from this
module, by calling :func:`request` (see below).

'''

from __future__ import unicode_literals, division

from ..helpers.misc import smartstr
from .classes import RegistryRegistry, RegistryBase

__all__ = ['request',
           ]


def request(name, autocreate=False, nodup=True, klass=None, *args, **kwargs):
    '''

    Returns a registry instance specified by ``name``, optionally creates
    one of type ``klass`` for you (with extra arguments passed as arguments
    to ``klass.__init__``), if no registry with name ``name`` exists.

    If the auto-creation behavior is turned off (which is the default), a
    :exc:`KeyError` will be ``raise``\ d if the requested instance does not
    exist. If auto-creation is enabled but ``klass`` is not derived from
    :class:`.classes.RegistryBase`, :exc:`TypeError` will be ``raise``\ d.

    To avoid getting an existing registry when you actually want a new one,
    a parameter ``nodup`` is added to indicate that any attempt to request
    a pre-existing registry with ``autocreate=True`` should result in an
    :exc:`AttributeError` (the same exception as a registry would ``raise``
    in such a situation). The parameter defaults to ``True``; it has no
    effect if ``autocreate == False``.

    '''

    # an extra layer of input type guarantee... is it needed?
    name = smartstr(name)

    if autocreate and nodup and name in _registries:
        raise AttributeError("the registry '%s' already exists" % name)

    try:
        return _registries[name]
    except KeyError:
        pass

    if not autocreate:
        raise KeyError("requested registry %s does not exist" %
                (name, )
                )

    # create one registry, but only instantiate klass iff klass is
    # RegistryBase-derived, to prevent unneeded object creation.
    if not issubclass(klass, RegistryBase):
        raise TypeError("class '%s' not subclass of RegistryBase" %
                (repr(klass), )
                )

    # instantiate, insert and return
    new_registry = klass(*args, **kwargs)
    _registries.register(name, new_registry)

    # in case of future expansion of registration semantics and possible
    # post-processing, return via the central registry singleton.
    return _registries[name]


# init at module load
_registries = RegistryRegistry()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / session handling / Beaker integration
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

import six

from beaker.session import Session, SessionObject

# we must do option processing ourselves
from beaker.util import coerce_session_params

from . import session_hub


class BeakerSession(object):
    def __init__(self, options=None):
        # defaults of Beaker's SessionMiddleware
        self.options = {
                'invalidate_corrupt': True,
                'type': None,
                'data_dir': None,
                'key': 'beaker.session.id',
                'timeout': None,
                'secret': None,
                'log_file': None,
                }

        if options is not None:
            # Simplify the filtering present in Beaker's implementation
            # This is also required to mimic the 'Assume all keys are
            # intended for cache if none are prefixed w/ "cache."' behavior
            for k, v in six.iteritems(options):
                if k.startswith('session.'):
                    self.options[k[8:]] = v

        coerce_session_params(self.options)

        if not self.options and options:
            self.options = options

    def preprocess(self, request):
        request.session = SessionObject(request.env, **self.options)

        return None

    def postprocess(self, response):
        request = response.request
        ctx = response.context
        session = request.session

        if session.accessed():
            session.persist()

            _headers = session.__dict__['_headers']
            if _headers['set_cookie']:
                cookie = _headers['cookie_out']
                if cookie:
                    # append the cookie to response
                    new_cookies = ctx.get('cookies', [])
                    new_cookies.append(cookie)
                    ctx['cookies'] = new_cookies

        return response


@session_hub.register_handler('beaker')
def beaker_session_handler(hub, options=None):
    return BeakerSession(options)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / Mako renderer
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Mako renderer
-------------

This renderer, and the associated handler for type ``'mako'``, provides
integration with the `Mako templating system`_. It reads configuration
from the key ``'mako'`` in registry ``'weiyu.rendering'``.

.. _Mako templating system: http://www.makotemplates.org/


Configuration options
^^^^^^^^^^^^^^^^^^^^^

Path to template directories can (and must) be configured. Also optionally
one may specify a directory for storing compiled template code.

+------------+-----------------+-------------------------------------+
|Config key  |Type of value    |Meaning                              |
+============+=================+=====================================+
|directories |list of strings  |Paths to template directories        |
+------------+-----------------+-------------------------------------+
|module_dir  |(unicode) string |Path to store compiled template code |
|            |                 |in, typically one in ``/tmp``        |
+------------+-----------------+-------------------------------------+

'''

from __future__ import unicode_literals, division

__all__ = ['MakoRenderable', ]

from os.path import abspath

from mako.lookup import TemplateLookup

from . import render_hub
from .base import Renderable

from ..registry.provider import request

TMPL_LOOKUP_KEY = 'lookup_obj'
DIRECTORIES_KEY = 'directories'
MODULE_DIR_KEY = 'module_dir'


class MakoRenderable(Renderable):
    def __init__(self, tmpl, *args, **kwargs):
        super(MakoRenderable, self).__init__(*args, **kwargs)
        self._template = tmpl

    def _do_render(self, result, context):
        real_ctx = result.copy()
        real_ctx.update(context)
        return self._template.render_unicode(**real_ctx), {}


@render_hub.register_handler('mako')
def mako_render_handler(hub, name, *args, **kwargs):
    render_reg = request('weiyu.rendering')
    # TODO: config default value here, or proper exc handling
    mako_params = render_reg['mako']
    if TMPL_LOOKUP_KEY not in mako_params:
        # canonicalize template dirs to absolute path
        abspath_dirs = [abspath(i) for i in mako_params[DIRECTORIES_KEY]]

        # instantiate TemplateLookup singleton
        mako_params[TMPL_LOOKUP_KEY] = TemplateLookup(
                directories=abspath_dirs,
                module_directory=mako_params[MODULE_DIR_KEY],
                )

    lookup = mako_params[TMPL_LOOKUP_KEY]
    # TODO: possibly map template names to filenames here
    tmpl = lookup.get_template(name)
    return MakoRenderable(tmpl, *args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / utilities / useful views
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

from os.path import exists, isdir, islink, normcase, sep
from os.path import normpath, abspath, realpath, join as pathjoin
from mimetypes import guess_type

from ..shortcuts import http, view


@http
@view
def staticfile_view(request, path):
    # TODO: caching
    conf = request.site['staticfile']
    STATIC_ROOT = normcase(conf['root'])
    #print 'path:', path
    #print 'root:', STATIC_ROOT

    # XXX WARNING From here SECURITY IS VERY IMPORTANT!!
    # first do NOT let ANY '..' things pass thru by forcing a normalization
    norm_relpath = normpath('/' + path).lstrip(sep)
    #print 'norm:', norm_relpath

    # concat the normalized thing and STATIC_ROOT together
    # XXX Here any potential symlink is a security hole! Production
    # environment should NEVER use this view; instead PLEASE rely on the
    # web server to do the job!
    real_path = realpath(abspath(pathjoin(STATIC_ROOT, norm_relpath)))
    real_path = normcase(real_path)
    #print 'real:', real_path

    if exists(real_path) and not isdir(real_path):
        # decide mimetype solely by using filename
        # mimetype is either a bytestring or None
        mimetype, enc_prog = guess_type(real_path)
        if mimetype is None:
            # fallback to binary stream
            mimetype = 'application/octet-stream'

        # we should prepare a file pointer
        fp = open(real_path, 'rb')

        # return a raw file response
        return (
                200,
                {'sendfile_fp': fp, },
                {'is_raw_file': True, 'mimetype': mimetype, },
                )

    # not found, return a 404
    # FIXME: This view is NOT renderable, this is SURE to raise exc later!
    return (
            404,
            {},
            {},
            )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / request service bundle / classes
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Reflex-like request handling
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the general request handling mechanism, much like
the neurological concept *reflex arc*, hence the name. Process of requests
and responses are fully customizable via the hooking machinery, which can
be done on the subclasses.

'''

from __future__ import unicode_literals, division

import abc

import six


class ReflexRequest(dict):
    '''Class representing a protocol-independent request.'''

    def __init__(self, env, *args, **kwargs):
        self.env = env


class ReflexResponse(dict):
    '''Class describing a protocol-independent response.'''

    def __init__(self, status, content, context=None, request=None):
        self.status = status
        self.content = content
        self.context = context or {}
        self.request = request


class BaseReflex(six.with_metaclass(abc.ABCMeta)):
    '''Abstract reflex class.

    Responses are obtained by calling :meth:`stimulate`, which "excites" the
    several ``_do_*`` methods in order. Among these methods
    :meth:`_do_accept_request`, :meth:`_do_generate_response` and
    :meth:`_do_deliver_response` are abstract; they are to implement the bare
    minimum of code to parse incoming requests into protocol-agnostic form,
    to generate a response, to transform and deliver the response object,
    thus subclasses must provide implementations.

    '''

    @abc.abstractmethod
    def _do_accept_request(self, *args, **kwargs):
        '''Called to convert protocol-specific request parameters into an
        instance of a certain subclass of :class:`.ReflexRequest`, for
        further processing.

        This method is abstract.

        '''

        pass

    def _do_translate_request(self, request):
        '''Called to perform various protocol-agnostic transformations on
        ``request``, such as applying middleware parameters or recording
        session information.

        The default implementation does nothing: it just returns the object
        untouched. Subclasses are free to override this.

        '''

        return request

    @abc.abstractmethod
    def _do_generate_response(self, request):
        '''Called to generate a :class:`.ReflexResponse` (sub)class instance
        with the information present in ``request``.

        The method is abstract by nature.

        '''

        pass

    def _do_postprocess(self, response):
        '''Called to perform protocol-agnostic transformations on
        ``response``.

        Same as with :meth:`_do_translate_request`, the default
        implementation is a stub which is free to be overridden.

        '''

        return response

    @abc.abstractmethod
    def _do_deliver_response(self, response):
        '''Called to actually deliver content in ``response`` in a
        particular output protocol.

        This method is also abstract by nature. There is no restrictions
        put on its return value; after the method is called the whole
        stimulation procedure ends, and you are in total control.

        '''

        pass

    def stimulate(self, *args, **kwargs):
        '''Triggers a response upon being presented with a set of arbitrary
        parameters. Depending on implementations of the several relaying
        methods, the effects triggered can be just about anything.

        '''

        raw_request = self._do_accept_request(*args, **kwargs)
        request = self._do_translate_request(raw_request)
        raw_response = self._do_generate_response(request)
        response = self._do_postprocess(raw_response)
        return self._do_deliver_response(response)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / utilities / server
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division, print_function

__all__ = [
        'cli_server',
        ]

import sys
import inspect
from socket import gethostname

from ..adapters import adapter_hub

DEFAULT_PORT = 9090
_SERVER_FLAVORS = {}


def expose_flavor(name):
    def _decorator_(thing):
        _SERVER_FLAVORS[name] = thing
        return thing
    return _decorator_


def get_port_number():
    if len(sys.argv) > 2:
        print(
                'usage: %s [port=%d]' % (sys.argv[0], DEFAULT_PORT),
                file=sys.stderr,
                )
        sys.exit(2)

    return int(sys.argv[1]) if len(sys.argv) == 2 else DEFAULT_PORT


def cli_server(flavor, *args, **kwargs):
    return _SERVER_FLAVORS[flavor](*args, **kwargs)


@expose_flavor('cherrypy')
def cli_server_cherrypy(application=None, port=None, hostname=None):
    try:
        from cherrypy import wsgiserver
    except ImportError:
        print(
                'no cherrypy, plz run via an external wsgi server',
                file=sys.stderr,
                )
        sys.exit(1)

    port = get_port_number() if port is None else port
    hostname = gethostname() if hostname is None else hostname

    if application is None:
        # inspect the outer frame's locals to get the application object, for
        # developer's convenience.
        outer_frame = inspect.getouterframes(inspect.currentframe())[2][0]
        app = outer_frame.f_globals['application']
    else:
        app = application

    server = wsgiserver.CherryPyWSGIServer(
            ('0.0.0.0', port),
            app,
            server_name=hostname,
            )

    server.start()


@expose_flavor('tornado')
def cli_server_tornado(server=None, port=None, hostname=None):
    try:
        from tornado import ioloop
    except ImportError:
        print(
                'import of tornado.ioloop failed, bailing',
                file=sys.stderr,
                )
        sys.exit(1)

    port = get_port_number() if port is None else port
    hostname = gethostname() if hostname is None else hostname

    if server is None:
        # NOTE: Tornado does not have the notion of "application" as WSGI.
        # The procedure to start serving requests is also different, so we
        # do not inspect the parent frame for the application object.
        # Instead, we request one from ``adapter_hub`` directly.
        server = adapter_hub.make_app('tornado')

    server.listen(port)
    ioloop.IOLoop.instance().start()


@expose_flavor('socketio')
def cli_server_socketio(listener, application=None, *args, **kwargs):
    try:
        from socketio.server import SocketIOServer
    except ImportError:
        print(
                'import of socketio.server failed, bailing',
                file=sys.stderr,
                )
        sys.exit(1)

    if application is None:
        # inspect the caller
        outer_frame = inspect.getouterframes(inspect.currentframe())[2][0]
        app = outer_frame.f_globals['application']
    else:
        app = application

    server = SocketIOServer(listener, app, *args, **kwargs)
    server.serve_forever()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / object-nonrelational mapper / base model
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'Document',
        ]

from functools import partial

from .. import db_hub
from . import mapper_hub

from ...helpers.metaprogramming import classproperty, classinstancemethod


class Document(dict):
    # only struct id is needed here, database association is done in
    # configuration file
    struct_id = None

    def __repr__(self):
        return b'<%s: %s>' % (
                self.struct_id,
                super(Document, self).__repr__(),
                )

    @classproperty
    def storage(cls):
        '''Get the driver-specific object for operating the underlying
        storage.

        This should be the collection object ready for operations, provided
        by the database driver.

        '''

        assert cls.struct_id is not None
        return db_hub.get_storage(cls.struct_id)

    @classinstancemethod
    def encode(self, cls, obj=None, version=None):
        '''Encode the document into a database-ready form.

        This method, and its companion :meth:`decode`, can be invoked both
        as class method and as instance method; if invoked from an instance,
        the parameter ``obj`` is ignored.

        '''

        assert cls.struct_id is not None
        return mapper_hub.encode(cls.struct_id, self or obj, version)

    @classinstancemethod
    def decode(self, cls, obj=None, version=None):
        assert cls.struct_id is not None
        return mapper_hub.decode(cls.struct_id, self or obj, version)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / cache drivers / redis driver
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

import redis

from ...db import db_hub
from .. import cache_hub
from .baseclass import BaseCache


class RedisCache(BaseCache):
    def __init__(self, storage):
        super(RedisCache, self).__init__()

        # XXX: Very primitive persistent connection impl
        # This relies on the Redis driver's internal state-lessness
        # The driver's start() and finish() methods will NOT be called
        self._conn = storage.driver.get_bucket(storage.bucket)

    def get(self, k):
        return self._conn.get(k)

    def set(self, k, v, timeout=None):
        return self._conn.set(k, v)

    def delete(self, k):
        return self._conn.delete(k)


@cache_hub.register_handler('redis')
def redis_handler(hub, opts):
    struct_id = opts['struct_id']
    storage = db_hub.get_storage(struct_id)

    return RedisCache(storage)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / JSON renderer
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
JSON renderer
-------------

.. todo::

    This part of documentation is yet to be written.

'''

from __future__ import unicode_literals, division

__all__ = ['JSONRenderable', ]

try:
    # favor esnme's ujson accelerated library
    from ujson import dumps
except ImportError:
    # fallback on the stdlib
    from json import dumps

    # use compact encoding, so construct a shim to keep the function
    # signature consistent
    from functools import partial
    dumps = partial(dumps, separators=(',', ':', ))

from . import render_hub
from .base import Renderable

# FIXME: this doesn't work for [lte IE 7], as a Save As dialog would pop up.
# We need to sniff the UA string and return text/javascript in that case.
JSON_COMMON_EXTRAS = {
        'mimetype': 'application/json',
        }


class JSONRenderable(Renderable):
    def _do_render(self, result, context):
        # only expose the desired result object
        return dumps(dict(result)), JSON_COMMON_EXTRAS


@render_hub.register_handler('json')
def json_render_handler(hub, name, *args, **kwargs):
    # template name is meaningless in JSON rendering, thus ignored.
    return JSONRenderable(*args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / package
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'router_hub',
        ]

import six

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry

# this does not cause circular import
from .config.parser import parse_config

PROBER = ModProber('weiyu.router', '%srouter')


class RouterHub(BaseHub):
    registry_name = 'weiyu.router'
    registry_class = UnicodeRegistry
    handlers_key = 'handlers'

    def __init__(self):
        super(RouterHub, self).__init__()

        if 'endpoints' not in self._reg:
            self._reg['endpoints'] = {}

        if 'routers' not in self._reg:
            self._reg['routers'] = {}

        if 'classes' not in self._reg:
            self._reg['classes'] = {}

        # cache the references
        self._routers = self._reg['routers']
        self._endpoints = self._reg['endpoints']
        self._classes = self._reg['classes']

    def endpoint(self, typ, name):
        '''decorator for registering routing end points.'''

        def _decorator_(fn):
            if typ not in self._endpoints:
                # this type not already registered, probably not inited yet
                # let's give it a sensible default
                self._endpoints[typ] = {}
            self._endpoints[typ][name] = fn
            return fn
        return _decorator_

    def register_router_class(self, name):
        '''Decorator to make a class available for routing.

        This decorator is mainly intended for internal use, to give short
        names to the router classes.

        '''

        def _decorator_(cls):
            if name in self._classes:
                raise ValueError(
                        'duplicate router register name: \'%s\'' % (name, )
                        )

            self._classes[name] = cls
            return cls
        return _decorator_

    def register_router(self, router):
        # keep a reference to the router
        typ = router.name
        if typ is None:
            raise ValueError(
                    'only named routers can be registered this way'
                    )

        self._routers[typ] = router

        # also reserve a slot in endpoints dict, if one is not already
        # set up
        if typ not in self._endpoints:
            self._endpoints[typ] = {}

        # register the router's dispatch method as handler, and we're done
        # here we'd use dry dispatch instead...
        # But first construct a shim removing the hub parameter...
        @self.register_handler(typ)
        def _routing_shim_(hub, *args, **kwargs):
            return router.dry_dispatch(*args, **kwargs)

    def dry_dispatch(self, typ, querystr, *args):
        # typically used with args=(request, ) inside the framework
        # TODO: is it really useful to allow passing kwargs also?
        return self.do_handling(typ, querystr, *args)

    def reverser_for(self, typ, __cache={}):
        try:
            return __cache[typ]
        except KeyError:
            pass

        router = self._routers[typ]
        map = router.reverse_map
        pat_cache = {}

        def reverser(endpoint, **kwargs):
            # look up the endpoint, memoized.
            try:
                pat_str, pat_vars = pat_cache[endpoint]
            except KeyError:
                # split the possibly scoped endpoint name into components
                try:
                    colon = endpoint.index(':')
                    scope, name = endpoint[:colon], endpoint[colon + 1:]
                except ValueError:
                    scope, name = '', endpoint

                # look up the scope first
                try:
                    scope_map = map[scope]
                except KeyError:
                    # TODO: a NoReverseMatch here, as Django did?
                    raise ValueError(
                            "No scope named '%s' in router type '%s'" % (
                                scope,
                                typ,
                                ))

                try:
                    pat_str, pat_vars = scope_map[name]
                except KeyError:
                    raise ValueError(
                            "No endpoint '%s' exposed in scope '%s' of "
                            "router type '%s'" % (
                                name,
                                scope,
                                typ,
                                ))

                # memoize the result
                pat_cache[endpoint] = (pat_str, pat_vars, )

            # verify the parameters
            given_vars = set(six.iterkeys(kwargs))
            if given_vars != pat_vars:
                # parameter mismatch
                raise ValueError('Parameter mismatch')

            # successful match, construct the result
            return pat_str % kwargs

        # memoize the reverser instance
        __cache[typ] = reverser

        return reverser

    def _do_init_router(
            self,
            typ,
            routing_rules,
            lvl,
            parent_info,
            _list_types=(list, tuple, ),
            ):
        # recursive algorithm, watch out d-:
        # typ is not really useful except checking against endpoint reg
        #
        # let's construct the desired target initializer out of the
        # pattern-to-(endpoint-or-router) list
        #
        # Attribute processing.
        attrib_list = routing_rules[0]
        inherited_renderer, scope = (
                parent_info['inherited_renderer'],
                parent_info['scope'],
                )
        include_path, cls_name = None, None

        if isinstance(attrib_list, _list_types):
            # Multiple attributes.
            # separate the router class from the others
            # the class spec is hardcoded to be the 1st attrib in the list
            cls_name = attrib_list[0]

            # Process the other attributes.
            for attrib in attrib_list[1:]:
                k, v = attrib.split('=', 1)

                if k == 'renderer':
                    # case: renderer=xxx
                    # record the renderer to inherit
                    inherited_renderer = v
                elif k == 'include':
                    # case: include=xxx
                    # it must appear as the only attribute, if used!
                    # throw an exception.
                    raise RuntimeError(
                            'include directive must appear on its own'
                            )
                elif k == 'scope':
                    # case: scope=xxx
                    scope = v
        else:
            # only one attribute.
            # it is either the router class spec, or an include directive
            if attrib_list.startswith('include='):
                k, v = attrib_list.split('=', 1)
                include_path = v
            else:
                cls_name = attrib_list

        # Process include.
        if include_path is not None:
            # There must be no other routing rules described... or we would
            # not know what to include.
            if len(routing_rules) > 1:
                raise RuntimeError(
                        'included router description should not contain '
                        'any rules'
                        )

            # The included file is then treated as a normal routing
            # description.
            return self.init_router_from_config(typ, include_path)

        # Load the requested (built-in) router class.
        #
        # Custom router classes should be registered before config load, like
        # before.
        #
        # TODO: allow registration of custom modules in modprober mechanism
        try:
            cls = self._classes[cls_name]
        except KeyError:
            try:
                PROBER.modprobe(cls_name)
            except ImportError:
                raise RuntimeError(
                        'request of unknown router class \'%s\'' % (
                            cls_name,
                            ),
                        )

            # Assume that module has actually registered the wanted class...
            assert cls_name in self._classes
            cls = self._classes[cls_name]

        # Process rules.
        result_rules = []
        for pattern, target_spec, extra_data in routing_rules[1:]:
            if isinstance(target_spec, _list_types):
                # this is a router... recursively construct a router out
                # of it
                my_info = {
                        'inherited_renderer': inherited_renderer,
                        'scope': scope,
                        }
                tgt = self._do_init_router(typ, target_spec, lvl + 1, my_info)
            elif isinstance(target_spec, six.string_types):
                # target is endpoint... check against endpoint registry
                # this is where typ is used
                # enforce a little bit of encoding requirement, to make
                # messed up encoding problem surface fast
                tgt = self._endpoints[typ][six.text_type(target_spec)]
            else:
                # unrecognized target specification, pass it thru as is
                tgt = target_spec

            # process special data
            if extra_data is not None:
                # render_in
                if extra_data.get('render_in', None) == 'inherit':
                    # inherit renderer from parent
                    extra_data['render_in'] = inherited_renderer
            else:
                extra_data = {}

            # remember the target spec for the URL reverser
            extra_data['target_spec'] = target_spec

            # add a rule
            result_rules.append((pattern, tgt, extra_data, ))

        # construct a XxxRouter object with the routing rules just created
        # if toplevel router, assign it a name equal to typ
        return cls(
                result_rules,
                name=typ if lvl == 0 else None,
                scope=scope,
                )

    def init_router(self, typ, routing_rules):
        return self._do_init_router(
                typ,
                routing_rules,
                0,
                {
                    'inherited_renderer': None,
                    'scope': '',
                    },
                )

    def init_router_from_config(self, typ, filename):
        config = parse_config(filename)
        return self.init_router(typ, config)


router_hub = RouterHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / utilities / GitHub webhook listener
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

try:
    import ujson as json
except ImportError:
    import json

try:
    import ipaddress
except ImportError:
    import ipaddr as ipaddress

from ..registry.provider import request as regrequest
from ..shortcuts import *

# Valid GitHub callback IP ranges
# TODO: according to GitHub, implement HTTP Basic auth as well
GH_IP_WHITELIST = (
        '204.232.175.64/27',
        '192.30.252.0/22',
        )

GH_NETWORKS = [ipaddress.IPv4Network(i) for i in GH_IP_WHITELIST]


def _dummy(status):
    return (
            status,
            None,
            {
                'mimetype': 'text/plain',
                },
            )


def is_ip_whitelisted(ip):
    addr = ipaddress.IPv4Address(ip)
    return any(addr in net for net in GH_NETWORKS)


@http('gh-webhook-post-receive')
@renderable('dummy')
@view
def on_gh_post_receive(request):
    repos = regrequest('site')['github']['post-receive']

    if not is_ip_whitelisted(request.remote_addr):
        return _dummy(403)

    if request.method != 'POST':
        # TODO: a limit method decorator would be better
        return _dummy(400)

    try:
        payload_json = request.form['payload']
    except KeyError:
        return _dummy(400)

    try:
        payload = json.loads(payload_json)
    except ValueError:
        return _dummy(400)

    repo_name, owner_name, ref = (
            payload['repository']['name'],
            payload['repository']['owner']['name'],
            payload['ref'],
            )

    # format config key
    if '/' in owner_name or '/' in repo_name:
        # Malformed names!
        return _dummy(400)

    conf_key = '%s/%s' % (owner_name, repo_name, )

    if conf_key not in repos:
        return _dummy(403)

    if ref not in repos[conf_key]:
        return _dummy(403)

    # Push accepted, execute the command given in configuration
    # Write the payload into the configured file, in effect also touching
    # it
    # FIXME: This is best done via some established deferred mechanism
    with open(repos[conf_key][ref], 'wb') as fp:
        fp.write(payload_json)

    return _dummy(204)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / db drivers / riak driver
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Riak Driver
~~~~~~~~~~~

This is the Riak driver for ``weiyu``.

'''

from __future__ import unicode_literals, division

import riak

from .. import db_hub
from .baseclass import BaseDriver


class RiakDriver(BaseDriver):
    def __init__(self, protocol, host, http_port, pb_port, nodes):
        super(RiakDriver, self).__init__()

        self.host, self.http_port, self.pb_port, self.nodes = (
                host,
                http_port,
                pb_port,
                nodes,
                )

        if nodes is None:
            # Let Riak client create a node for us
            # Parameters end up in RiakClient.__init__'s **unused_args,
            # which in turn are passed on to RiakNode's ctor.
            client = riak.RiakClient(
                    protocol=protocol,
                    host=host,
                    http_port=http_port,
                    pb_port=pb_port,
                    )
        else:
            # host, http_port and pb_port are ignored if a list of node
            # configurations is specified
            # TODO: provide way to config transport_options?
            client = riak.RiakClient(
                    protocol=protocol,
                    nodes=nodes,
                    )

        self.conn = client
        self._buckets = {}

    def start(self):
        # No-op.
        pass

    def finish(self):
        # No-op.
        pass

    def get_bucket(self, bucket):
        try:
            return self._buckets[bucket]
        except KeyError:
            b = self.conn.bucket(bucket)
            self._buckets[bucket] = b

        return self._buckets[bucket]

    def __repr__(self):
        return b'<weiyu.db/riak: %s:%d, prefix=%s, mapred_prefix=%s>' % (
                self.host,
                self.port,
                self.prefix,
                self.mapred_prefix,
                )


@db_hub.register_handler('riak')
def riak_handler(
        hub,
        protocol='http',
        host='127.0.0.1',
        http_port=8098,
        pb_port=8087,
        nodes=None,
        ):
    return RiakDriver(protocol, host, http_port, pb_port, nodes)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / db / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'db_hub',
        'mapper_hub',
        ]

import six

from ..helpers.hub import BaseHub
from ..helpers.misc import smartstr
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry
from ..registry.provider import request

DBCONF_KEY, DRVOBJ_KEY = 'databases', 'drvobjs'
STORAGE_KEY, STORAGE_CACHE_KEY = 'storage', '_storage_cache'

PROBER = ModProber('weiyu.db.drivers', '%s_driver')


class DatabaseHub(BaseHub):
    registry_name = 'weiyu.db'
    registry_class = UnicodeRegistry
    handlers_key = 'drivers'

    def __init__(self, *args, **kwargs):
        super(DatabaseHub, self).__init__(*args, **kwargs)

        self._drvobjs = self._reg[DRVOBJ_KEY] = {}
        self._storage_cache = self._reg[STORAGE_CACHE_KEY] = {}

    # This association MUST be done AFTER central registry is fully
    # populated with user config, or the registry mechanism will
    # prevent user config values from being loaded.
    def _init_refresh_map(self):
        # register an empty database dict if it isn't there
        if DBCONF_KEY not in self._reg:
            self._reg[DBCONF_KEY] = {}

        # and storage configuration
        if STORAGE_KEY not in self._reg:
            self._reg[STORAGE_KEY] = {}
        self._storage = self._reg[STORAGE_KEY]

    def get_database(self, name):
        # only fetch resource by name
        # NOTE: args and kwargs are removed for the moment; we'll see if
        # they're really necessary for the "flexibility" they're supposed
        # to provide.
        try:
            return self._drvobjs[name]
        except KeyError:
            drv = self.do_handling('__name', name)
            self._drvobjs[name] = drv
            return drv

    def get_storage_conf(self, name):
        # First let's do a cached lookup...
        try:
            return self._storage_cache[name]
        except KeyError:
            pass

        try:
            cfg = self._storage[name]
        except KeyError:
            raise TypeError(
            "struct id '%s' does not have storage configured" % name
            )

        if isinstance(cfg, dict):
            # custom config, pass as-is
            cfg_dict = cfg
        elif isinstance(cfg, six.string_types):
            # shorthand for the vast majority of db/bucket setup
            # the config dict is created here
            if not isinstance(cfg, six.text_type):
                cfg = cfg.decode('utf-8')

            db, bucket = cfg.split('/', 1)

            # coerce bucket param, needed e.g. by Redis
            if bucket.startswith('(int)'):
                bucket = int(bucket[5:])

            cfg_dict = {'db': db, 'bucket': bucket, }
        else:
            raise ValueError(
            "struct id '%s': Don't know how to handle config %s" % (
                name,
                repr(cfg),
                )
            )

        self._storage_cache[name] = cfg_dict
        return cfg_dict

    def get_storage(self, name):
        storage_conf = self.get_storage_conf(name)
        db_name, bucket = storage_conf['db'], storage_conf['bucket']
        drv = self.get_database(db_name)

        return drv(bucket)


db_hub = DatabaseHub()


# reference-by-name handler
# driver objects are memoized in db_hub.get_database()
@db_hub.register_handler('__name')
def name_resolver(hub, name):
    dbconf = request('weiyu.db')
    # NOTE: exception is not caught as any request for an unmentioned
    # database SHOULD fail
    db_cfg = dbconf[DBCONF_KEY][smartstr(name)]

    # driver type and kwargs for constructing db object
    drv_type = db_cfg.pop('driver')

    # import driver module
    PROBER.modprobe(drv_type)

    return hub.do_handling(drv_type, **db_cfg)


# expose document mapper here to prevent circular dependency
from .mapper import mapper_hub


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / cache drivers / base class
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

import abc

import six


class BaseCache(six.with_metaclass(abc.ABCMeta)):
    def __init__(self):
        super(BaseCache, self).__init__()

    @abc.abstractmethod
    def get(self, k):
        return None

    @abc.abstractmethod
    def set(self, k, v, timeout=None):
        return True

    @abc.abstractmethod
    def delete(self, k):
        return True

    def __getitem__(self, k):
        return self.get(k)

    def __setitem__(self, k, v):
        return self.set(k, v)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / SCSS renderer
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
SCSS renderer
-------------

This renderer is a wrapper around the `pyScss`_ SCSS library. The renderer
reads its configuration from the key ``'scss'`` in registry
``'weiyu.rendering'``.

.. _pyScss: https://github.com/Kronuz/pyScss

'''

from __future__ import unicode_literals, division

__all__ = ['SCSSRenderable', ]

from os.path import abspath

import scss

from . import render_hub
from .base import Renderable

from ..registry.provider import request

SCSS_COMMON_EXTRAS = {
        'mimetype': 'text/css',
        }

SCSS_CONFIG_KEYS = (
        'STATIC_ROOT STATIC_URL ASSETS_ROOT ASSETS_URL LOAD_PATHS'
        ).split(' ')

SCSS_OPTS_KEY = 'opts'
SCSS_DEFAULT_OPTS = {
        'compress': True,
        }


class SCSSRenderable(Renderable):
    def __init__(self, options, *args, **kwargs):
        super(SCSSRenderable, self).__init__(*args, **kwargs)
        self._options = options

    def _do_render(self, result, context):
        # Filename is passed via context due to limitation of @renderable(),
        # which is primarily designed for templating systems.
        filename = context['scss_file']

        # Compiler.
        # result is the context variables of SCSS
        compiler = scss.Scss(scss_vars=result, scss_opts=self._options)

        return compiler.compile(scss_file=filename), SCSS_COMMON_EXTRAS


@render_hub.register_handler('scss')
def scss_render_handler(hub, name, *args, **kwargs):
    render_reg = request('weiyu.rendering')
    # TODO: default value
    scss_params = render_reg['scss']

    if '_inited' not in scss_params:
        # config global params
        for k in SCSS_CONFIG_KEYS:
            if k in scss_params:
                setattr(scss.config, k, scss_params[k])

        # prepare compile options
        options = scss_params['_options'] = SCSS_DEFAULT_OPTS.copy()
        if SCSS_OPTS_KEY in scss_params:
            options.update(scss_params[SCSS_OPTS_KEY])

        scss_params['_inited'] = True

    options = scss_params['_options']

    return SCSSRenderable(options, *args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / db drivers / redis driver
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Redis Driver
~~~~~~~~~~~

This is the Redis driver for ``weiyu``.

Note that the "db ordinal" is used as bucket names, and not specified
in the connection parameters.

'''

from __future__ import unicode_literals, division

import redis

from .. import db_hub
from .baseclass import BaseDriver


class RedisDriver(BaseDriver):
    def __init__(self, host, port):
        super(RedisDriver, self).__init__()

        self.host, self.port = host, port
        self._buckets = {}

    def start(self):
        # No-op.
        pass

    def finish(self):
        # No-op.
        pass

    def get_bucket(self, bucket):
        try:
            return self._buckets[bucket]
        except KeyError:
            b = redis.StrictRedis(self.host, self.port, bucket)
            self._buckets[bucket] = b

        return self._buckets[bucket]

    def __repr__(self):
        return b'<weiyu.db/redis: %s:%d>' % (self.host, self.port, )


@db_hub.register_handler('redis')
def redis_handler(
        hub,
        host='127.0.0.1',
        port=6379,
        ):
    return RedisDriver(host, port)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / bridging views
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'scss_bridge_view',
        ]

from os.path import exists, normcase, normpath, realpath, abspath, sep
from os.path import join as pathjoin

from .decorators import renderable
from ..shortcuts import http, view


@http
@renderable('scss')
@view
def scss_bridge_view(request, path):
    conf = request.site['scss-bridge']
    SCSS_ROOT = normcase(conf['root'])

    # TODO: extract this path manipulation code out (which is the same as
    # that in staticfile_view)
    norm_relpath = normpath('/' + path).lstrip(sep)
    real_path = normcase(realpath(abspath(pathjoin(SCSS_ROOT, norm_relpath))))

    if not exists(real_path):
        return (
                404,
                {},
                {},
                )

    return (
            200,
            {},
            {
                'scss_file': real_path,
                },
            )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / base operations
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division


class RenderingError(Exception):
    pass


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / signals / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'signal_hub',
        ]

from ..helpers.hub import BaseHub
from ..registry.classes import UnicodeRegistry

LISTENERS_KEY = 'listeners'


class SignalHub(BaseHub):
    registry_name = 'weiyu.signal'
    registry_class = UnicodeRegistry
    handlers_key = 'handlers'  # XXX Not used!

    def __init__(self):
        super(SignalHub, self).__init__()

        if LISTENERS_KEY not in self._reg:
            self._reg[LISTENERS_KEY] = {}

        self._listeners = self._reg[LISTENERS_KEY]

    def append_listener_to(self, name):
        def _decorator_(thing):
            if name in self._listeners:
                self._listeners[name].append(thing)
            else:
                self._listeners[name] = [thing, ]

            return thing
        return _decorator_

    def prepend_listener_to(self, name):
        def _decorator_(thing):
            if name in self._listeners:
                self._listeners[name].insert(0, thing)
            else:
                self._listeners[name] = [thing, ]

            return thing
        return _decorator_

    def fire(self, name, *args, **kwargs):
        try:
            callbacks = self._listeners[name]
        except KeyError:
            raise KeyError(
                    "signal '%s' not claimed by any listener" % (
                        name,
                        ))

        for fn in callbacks:
            fn(*args, **kwargs)

    def fire_nullok(self, name, *args, **kwargs):
        if name not in self._listeners:
            return
        return self.fire(name, *args, **kwargs)


signal_hub = SignalHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / framework-level initialization / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'load_router',
        'load_config',
        'load_views',
        'boot',
        ]

import inspect
import warnings

import six

from ..adapters import adapter_hub
from ..db import db_hub
from ..router import router_hub
from ..registry.classes import UnicodeRegistry
from ..registry.loader import BaseConfig
from ..registry.provider import request as regrequest
from .viewloader import ViewLoader

# XXX Force load of HTTPSessionMiddleware
from ..adapters.http import base
del base


def load_router(typ, filename):
    router = router_hub.init_router_from_config(typ, filename)
    return router_hub.register_router(router)


def load_config(path):
    ret = BaseConfig.get_config(path).populate_central_regs()

    # Refresh the hubs' internal cached references, this MUST be done
    # after config has loaded.
    db_hub._init_refresh_map()

    return ret


def load_views(path_or_config):
    if isinstance(path_or_config, six.string_types):
        # this is a path to views.json
        return ViewLoader().fileconfig(path_or_config)()

    # direct config by dict
    return ViewLoader(path_or_config)()


def boot(
        conf_path='conf.yml',
        views_path=None,
        router_type='http',
        root_router_file=None,
        ):
    # initialize registries, views, router, in that order
    load_config(conf_path)

    # make an empty site registry if not present
    reg_site = regrequest(
            'site',
            autocreate=True,
            nodup=False,
            klass=UnicodeRegistry,
            )

    # determine if ViewLoader is configured by registry
    if views_path is not None:
        # external file configured by app stub
        load_views(views_path)
    else:
        # extract config from site registry
        if 'views' in reg_site:
            view_config = reg_site['views']
            load_views(view_config)
        else:
            # old behavior, which is an implicit (specified via default)
            # views.json, kept for compatibility
            warnings.warn(
                    'Implicit views.json is discouraged, please put the '
                    'view config directly into main config file',
                    PendingDeprecationWarning,
                    )
            load_views('views.json')


    # init router
    if root_router_file is not None:
        # configured by app stub
        router_file_to_use = root_router_file
    else:
        if 'urls' in reg_site:
            # configured by registry
            router_file_to_use = reg_site['urls']
        else:
            # default value to use
            # this is not deprecated because urls.txt has special syntax
            # anyway
            router_file_to_use = 'urls.txt'

    load_router(router_type, router_file_to_use)

    # register middlewares according to config
    middleware_decl = (
            reg_site['middlewares']
            if 'middlewares' in reg_site
            else {}
            )

    def call_register_middleware(kind):
        middlewares = middleware_decl.get(kind, [])
        adapter_hub.register_middleware_chain(
                router_type,
                kind,
                middlewares,
                )

    call_register_middleware('pre')
    call_register_middleware('post')


def inject_app(app_type='wsgi', var='application', *args, **kwargs):
    # Make app
    boot(*args, **kwargs)
    app = adapter_hub.make_app(app_type)

    # Then inject the object into the caller's global namespace
    parentframe = inspect.getouterframes(inspect.currentframe())[1][0]
    parentframe.f_globals[var] = app


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'adapter_hub',
        ]

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry
from ..signals import signal_hub

ADAPTERS_KEY = 'adapters'
KNOWN_MIDDLEWARE_KEY = 'known_middlewares'
USED_MIDDLEWARE_KEY = 'used_middlewares'

PROBER = ModProber(
        'weiyu.adapters',
        '%s',
        {
            'wsgi': 'http.wsgi',
            'tornado': 'http.tornado_',
            },
        )


class AdapterHub(BaseHub):
    registry_name = 'weiyu.adapter'
    registry_class = UnicodeRegistry
    handlers_key = ADAPTERS_KEY

    def __init__(self):
        super(AdapterHub, self).__init__()
        self._middlewares = self._reg[KNOWN_MIDDLEWARE_KEY] = {}
        self._middleware_in_use = self._reg[USED_MIDDLEWARE_KEY] = {}

    def make_app(self, adapter):
        PROBER.modprobe(adapter)
        return self.do_handling(adapter)

    def declare_middleware(self, name):
        def _decorator_(obj):
            if name in self._middlewares:
                raise ValueError(
                        "middleware '%s' already declared" % (
                            name,
                            )
                        )

            self._middlewares[name] = obj
            return obj
        return _decorator_

    def register_middleware_chain(self, router_type, chain, middleware_names):
        signal_name = '%s-middleware-%s' % (router_type, chain, )

        for name in middleware_names:
            # Instantiate each middleware object only once
            try:
                middleware = self._middleware_in_use[name]
            except KeyError:
                middleware_cls = self._middlewares[name]
                middleware = self._middleware_in_use[name] = middleware_cls()

            method = getattr(middleware, 'do_%s' % (chain, ))
            signal_hub.append_listener_to(signal_name)(method)


adapter_hub = AdapterHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:
